# -*- coding: utf-8 -*-

"""
Unit tests for the normal form detection.

Author: Gertjan van den Burg

"""

import unittest

from clevercsv.dialect import SimpleDialect
from clevercsv.normal_form import is_form_1
from clevercsv.normal_form import is_form_2
from clevercsv.normal_form import is_form_3
from clevercsv.normal_form import is_form_4
from clevercsv.normal_form import is_form_5


class NormalFormTestCase(unittest.TestCase):
    def test_form_1(self) -> None:
        dialect = SimpleDialect(delimiter=",", quotechar='"', escapechar="")

        self.assertTrue(is_form_1('"A","B","C"'.split("\n"), dialect))
        self.assertTrue(is_form_1('"A","B"\n"C","D"'.split("\n"), dialect))
        self.assertTrue(is_form_1('"A","","C"'.split("\n"), dialect))

        self.assertFalse(is_form_1('"A","B"\n"A"'.split("\n"), dialect))
        self.assertFalse(is_form_1('"A"\n"B"'.split("\n"), dialect))
        self.assertFalse(is_form_1('"A"\n"A","B"'.split("\n"), dialect))
        self.assertFalse(is_form_1('"A",,"C"'.split("\n"), dialect))
        self.assertFalse(is_form_1('"A",C'.split("\n"), dialect))
        self.assertFalse(is_form_1('"A"\n"b""A""c","B"'.split("\n"), dialect))

    def test_form_2(self) -> None:
        dialect = SimpleDialect(delimiter=",", quotechar="", escapechar="")

        self.assertTrue(is_form_2("1,2,3".split("\n"), dialect))
        self.assertTrue(is_form_2("1,2,3\na,b,c".split("\n"), dialect))
        self.assertTrue(is_form_2("a@b.com,3".split("\n"), dialect))
        self.assertTrue(is_form_2("a,,3\n1,2,3".split("\n"), dialect))

        self.assertFalse(is_form_2("1,2,3\n1,2\n4,5,6".split("\n"), dialect))
        self.assertFalse(is_form_2("1".split("\n"), dialect))
        self.assertFalse(is_form_2('1,"a"'.split("\n"), dialect))
        self.assertFalse(is_form_2("a;b,3".split("\n"), dialect))
        self.assertFalse(is_form_2('"a,3,3\n1,2,3'.split("\n"), dialect))
        self.assertFalse(is_form_2('a,"",3\n1,2,3'.split("\n"), dialect))

    def test_form_3(self) -> None:
        A = SimpleDialect(delimiter=",", quotechar="'", escapechar="")
        Q = SimpleDialect(delimiter=",", quotechar='"', escapechar="")

        self.assertTrue(is_form_3('A,B\nC,"D"'.split("\n"), Q))
        self.assertTrue(is_form_3('A,B\nC,"d,e"'.split("\n"), Q))

        self.assertFalse(is_form_3('A,\nC,"d,e"'.split("\n"), Q))
        self.assertFalse(is_form_3("3;4,B\nC,D".split("\n"), Q))

        self.assertFalse(is_form_3('A,B\n"C",D'.split("\n"), A))
        self.assertTrue(is_form_3('A,B\n"C",D'.split("\n"), Q))

    def test_form_4(self) -> None:
        quoted = SimpleDialect(delimiter="", quotechar='"', escapechar="")
        unquoted = SimpleDialect(delimiter="", quotechar="", escapechar="")

        self.assertTrue(is_form_4("A\nB\nC".split("\n"), unquoted))
        self.assertTrue(is_form_4("1\n2\n3".split("\n"), unquoted))
        self.assertTrue(is_form_4("A_B\n1\n2".split("\n"), unquoted))
        self.assertTrue(is_form_4("A&B\n1\n2".split("\n"), unquoted))
        self.assertTrue(is_form_4("A&B\n-1\n2".split("\n"), unquoted))
        self.assertTrue(is_form_4('"A"\n"B"\n"C"'.split("\n"), quoted))

        self.assertFalse(is_form_4('"A", "B"\n"B"\n"C"'.split("\n"), quoted))
        self.assertFalse(is_form_4('"A","B"\n"B"\n"C"'.split("\n"), quoted))
        self.assertFalse(is_form_4('"A@b"\n"B"\n"C"'.split("\n"), quoted))
        self.assertFalse(is_form_4('A\n"-1"\n2'.split("\n"), unquoted))
        self.assertFalse(is_form_4("A B\n-1 3\n2 4".split("\n"), unquoted))

    def test_form_5(self) -> None:
        dialect = SimpleDialect(delimiter=",", quotechar='"', escapechar="")

        self.assertTrue(is_form_5('"A,B"\n"1,2"\n"3,4"'.split("\n"), dialect))
        self.assertTrue(is_form_5('"A,B"\n"1,"\n"2,3"'.split("\n"), dialect))

        self.assertFalse(is_form_5("A,B\n1,2\n3,4".split("\n"), dialect))
        self.assertFalse(is_form_5("A,B\n1,\n2,3".split("\n"), dialect))
        self.assertFalse(
            is_form_5('"A,""B"""\n"1,"\n"2,3"'.split("\n"), dialect)
        )


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python

"""
This script should be opened within tmux and no other tmux sessions should be 
running.

Author: Gertjan van den Burg

"""

import argparse
import codecs
import gzip
import io
import json
import os
import time
import unicodedata

import chardet
import libtmux


def get_encoding(filename):
    detector = chardet.UniversalDetector()
    final_chunk = False
    blk_size = 65536
    with open(filename, "rb") as fid:
        while (not final_chunk) and (not detector.done):
            chunk = fid.read(blk_size)
            if len(chunk) < blk_size:
                final_chunk = True
            detector.feed(chunk)
    detector.close()
    encoding = detector.result.get("encoding", None)
    return encoding


def load_file(filename, encoding="unknown"):
    if encoding == "unknown":
        encoding = get_encoding(filename)
    with open(filename, "r", newline="", encoding=encoding) as fid:
        try:
            return fid.read()
        except UnicodeDecodeError:
            print(
                "UnicodeDecodeError occurred for file: %s. "
                "This means the encoding was determined incorrectly "
                "or the file is corrupt." % filename
            )
            return None


def is_potential_escapechar(char, encoding):
    as_unicode = codecs.decode(bytes(char, encoding), encoding=encoding)
    ctr = unicodedata.category(as_unicode)
    block = ["!", "?", '"', "'", ".", ",", ";", ":", "%", "*", "&", "#", "/"]
    if ctr == "Po":
        if as_unicode in block:
            return False
        return True
    return False


def get_quotechar_options(data):
    if data is None:
        return set(["q", "a", "b", "t", "n"])
    options = set()
    for x in data:
        if x == '"':
            options.add("q")
        elif x == "'":
            options.add("a")
        elif x == "`":
            options.add("b")
        elif x == "~":
            options.add("t")
    options.add("n")
    return options


def has_initial_space(data, delim):
    if data is None:
        return True
    return "%s " % delim in data


def pairwise(iterable):
    "s - > (s0, s1), (s1, s2), (s2, s3), ..."
    a = iter(iterable)
    b = iter(iterable)
    next(b, None)
    return zip(a, b)


def get_escapechar_options(data, encoding, delim, quotechar):
    if data is None or encoding is None:
        return set(["\\"])
    escapes = set()
    for u, v in pairwise(data):
        if not is_potential_escapechar(u, encoding):
            continue
        if v in [delim, quotechar] and not u in [delim, quotechar]:
            escapes.add(u)
    return escapes


class Asker(object):
    def __init__(self, filename, pane):
        self.filename = filename
        self.opened_vim = False
        self.opened_less = False
        self.pane = pane
        self.note = None
        self.dialect = {
            "delimiter": None,
            "quotechar": None,
            "escapechar": None,
            "skipinitialspace": None,
            "header_idx": None,
        }
        self.skip = False
        self.encoding = None
        self.data = None
        self.decompressed_file = None

        if filename.endswith(".gz"):
            self.decompressed_file = os.path.splitext(filename)[0]
            with open(filename, "rb") as fid:
                with open(self.decompressed_file, "wb") as oid:
                    oid.write(gzip.decompress(fid.read()))
            self.filename = self.decompressed_file

    def load_file(self):
        self.encoding = get_encoding(self.filename)
        self.data = load_file(self.filename, encoding=self.encoding)

    def ask_question(
        self, prompt, key, options=None, highlight_char=None, max_length=1
    ):
        if not self.opened_vim and not self.opened_less:
            self.open_less()
        self.highlight_char(highlight_char)
        prompt = prompt + " " if not prompt.endswith(" ") else prompt
        while True:
            if options is None:
                ans = input(prompt)
            else:
                ans = input("%s[%s] " % (prompt, "/".join(sorted(options))))
            if ans == "quit":
                self.quit()
            elif ans == "less":
                self.close(clean=False)
                self.open_less()
            elif ans in ["vi", "vim"]:
                self.close(clean=False)
                self.open_vim()
            elif ans in ["hlt", "hltab"]:
                self.pane.send_keys("/\\t")
            elif ans in ["hls", "hlspace"]:
                self.pane.send_keys("/\\ ")
            elif ans == "skip":
                self.skip = True
                break
            elif ans == "note":
                self.note = input("Enter note: ").strip()
            elif ans == "none" or ans == "no":
                self.dialect[key] = None
                break
            elif ans == "\\t":
                self.dialect[key] = "\t"
                break
            elif len(ans.strip()) > max_length:
                print("Only length %i answers are allowed." % max_length)
            elif (not options is None) and (not ans.rstrip("\n") in options):
                print("Only these options are allowed: %r" % options)
            else:
                self.dialect[key] = ans.rstrip("\n")
                break

    def process(self):
        print("Annotating file: %s" % self.filename)
        self.ask_question("What is the delimiter?", "delimiter")
        if self.skip:
            return None

        self.load_file()
        if has_initial_space(self.data, self.dialect["delimiter"]):
            hval = "%s " % self.dialect["delimiter"]
            self.ask_question(
                "What is skipinitialspace?",
                "skipinitialspace",
                options=["t", "f"],
                highlight_char=hval,
            )
            if self.skip:
                return None
            self.dialect["skipinitialspace"] = {"t": True, "f": False}[
                self.dialect["skipinitialspace"]
            ]
        else:
            self.dialect["skipinitialspace"] = False

        q_options = get_quotechar_options(self.data)
        if list(q_options) == ["n"]:
            self.dialect["quotechar"] = None
        else:
            q_options.remove("n")
            hval = None
            if "q" in q_options:
                hval = '"'
            elif "a" in q_options:
                hval = "'"
            self.ask_question(
                "What is the quotation mark",
                "quotechar",
                options=q_options,
                highlight_char=hval,
            )
            if not self.dialect["quotechar"] is None:
                self.dialect["quotechar"] = {
                    "Q": '"',
                    "A": "'",
                    "B": "`",
                    "T": "~",
                }[self.dialect["quotechar"].upper()]
        if self.skip:
            return None

        e_options = get_escapechar_options(
            self.data,
            self.encoding,
            self.dialect["delimiter"],
            self.dialect["quotechar"],
        )
        if "n" in e_options:
            raise ValueError("'n' shouldn't be an option in escapechars")
        if len(e_options) == 1 and "\\" in e_options:
            hval = "\\\\"
        else:
            hval = None
        if e_options:
            self.ask_question(
                "What is the escape character?",
                "escapechar",
                options=e_options,
                highlight_char=hval,
            )
        if self.skip:
            return None

        n_lines = sum((1 for l in io.StringIO(self.data)))
        self.pane.send_keys("gg", enter=False, suppress_history=False)
        self.ask_question(
            "Where is the header?", "header_idx", max_length=max(1, n_lines)
        )
        if not self.dialect["header_idx"] is None:
            self.dialect["header_idx"] = int(self.dialect["header_idx"])
        if self.skip:
            return None

        if self.dialect["delimiter"] is None:
            self.dialect["delimiter"] = ""
        if self.dialect["quotechar"] is None:
            self.dialect["quotechar"] = ""
        if self.dialect["escapechar"] is None:
            self.dialect["escapechar"] = ""

        return True

    def highlight_char(self, char=None):
        if char is None:
            return
        self.pane.send_keys("/%s" % char)
        self.pane.send_keys("gg", enter=False, suppress_history=False)
        self.pane.send_keys("n", enter=False, suppress_history=False)

    def open_vim(self):
        self.pane.send_keys("vim %s" % self.filename)
        self.opened_vim = True

    def open_less(self):
        self.pane.send_keys("less -N -f %s" % self.filename)
        self.opened_less = True

    def close(self, clean=True):
        if self.opened_less:
            self.close_less()
        elif self.opened_vim:
            self.close_vim()
        self.pane.clear()
        if clean and not self.decompressed_file is None:
            os.unlink(self.decompressed_file)

    def close_vim(self):
        self.pane.send_keys(":q", suppress_history=False)
        time.sleep(0.5)
        self.opened_vim = False

    def close_less(self):
        self.pane.send_keys("q", suppress_history=False)
        time.sleep(0.1)
        self.opened_less = False

    def quit(self):
        self.close()
        self.pane.send_keys("exit")
        print("Thank you.")
        raise SystemExit


def annotate_file(filename, name, tmux_pane):
    print()
    asker = Asker(filename, tmux_pane)
    res = asker.process()
    out = {"filename": name + ".csv", "status": "ok"}
    if not asker.note is None:
        out["note"] = asker.note

    if res is None:
        out["status"] = "skip"
        asker.close()
        return out
    out["dialect"] = asker.dialect
    out["encoding"] = asker.encoding
    asker.close()
    return out


def dump_result(output_file, res):
    with open(output_file, "a") as fid:
        json.dump(res, fid, indent=2, sort_keys=True)


def init_tmux():
    tmux_server = libtmux.Server()
    tmux_sess = tmux_server.list_sessions()[-1]
    tmux_win = tmux_sess.attached_window
    less_pane = tmux_win.split_window(attach=False)

    return less_pane


def process(input_dir, output_dir):
    input_files = os.listdir(input_dir)
    input_files.sort()
    input_paths = [os.path.join(input_dir, f) for f in input_files]
    output_files = os.listdir(output_dir)
    output_paths = [os.path.join(output_dir, f) for f in output_files]

    data = []
    for ipath in input_paths:
        base, ext = os.path.splitext(os.path.basename(ipath))
        md5, base_ext = os.path.splitext(base)
        if not (ext == ".csv" or (ext == ".gz" and base_ext == ".csv")):
            print("Warning: non-csv file found in input directory: %s" % ipath)
            continue
        opath = os.path.join(output_dir, md5 + ".json")
        if opath in output_paths:
            # result file exists
            continue
        name = base if base_ext == "" else os.path.splitext(base)[0]
        data.append((ipath, opath, name))

    if not data:
        return

    less_pane = init_tmux()
    start_time = time.time()

    count = 0
    for ipath, opath, name in data:
        res = annotate_file(ipath, name, less_pane)
        dump_result(opath, res)
        count += 1

        if count % 10 == 0:
            print(
                "\nProgress: %i done out of %i. "
                "This session: %i (%.2f seconds per file)"
                % (
                    count,
                    len(data),
                    count,
                    ((time.time() - start_time) / count),
                )
            )


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("input_dir", help="Input directory of csv files")
    parser.add_argument("output_dir", help="Output directory of json files")
    return parser.parse_args()


def main():
    args = parse_args()
    process(args.input_dir, args.output_dir)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

"""
Caller for the command line application.

"""

import sys

from ._optional import import_optional_dependency


def main() -> None:
    # Check that necessary dependencies are available
    import_optional_dependency("wilderness")

    # if so, load the actual main function and call it.
    from .console import main as realmain

    sys.exit(realmain())


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

from typing import Any
from typing import List
from typing import Optional

from clevercsv import __version__
from clevercsv.dialect import SimpleDialect


def parse_int(val: Any, name: str) -> Optional[int]:
    """Parse a number to an integer if possible"""
    if val is None:
        return val
    try:
        return int(val)
    except ValueError:
        raise ValueError(
            f"Please provide a number for {name}, instead of {val}"
        )


def generate_code(
    filename: str,
    dialect: SimpleDialect,
    encoding: Optional[str],
    use_pandas: bool = False,
) -> List[str]:
    assert dialect.quotechar is not None
    d = '"\\t"' if dialect.delimiter == "\t" else f'"{dialect.delimiter}"'
    q = '"%s"' % (dialect.quotechar.replace('"', '\\"'))
    e = repr(f"{dialect.escapechar}").replace("'", '"')
    base = [
        "",
        f"# Code generated with CleverCSV version {__version__}",
        "",
        "import clevercsv",
    ]
    if use_pandas:
        return [
            *base,
            "",
            f'df = clevercsv.read_dataframe("{filename}", delimiter={d}, '
            f"quotechar={q}, escapechar={e})",
            "",
        ]

    enc = "None" if encoding is None else f'"{encoding}"'
    lines = [
        *base,
        "",
        f'with open("{filename}", "r", newline="", encoding={enc}) as fp:',
        "    reader = clevercsv.reader(fp, "
        + f"delimiter={d}, quotechar={q}, escapechar={e})",
        "    rows = list(reader)",
        "",
    ]
    return lines



# -*- coding: utf-8 -*-

from __future__ import annotations

import csv
import os
import sys

from typing import TYPE_CHECKING
from typing import Any
from typing import Mapping
from typing import Type
from typing import TypeVar
from typing import Union

from clevercsv.dialect import SimpleDialect

AnyPath = Union[str, bytes, "os.PathLike[str]", "os.PathLike[bytes]"]
StrPath = Union[str, "os.PathLike[str]"]
_OpenFile = Union[AnyPath, int]
_DictRow = Mapping[str, Any]
_DialectLike = Union[str, csv.Dialect, Type[csv.Dialect], SimpleDialect]
_T = TypeVar("_T")

if sys.version_info >= (3, 8):
    from typing import Dict as _DictReadMapping
else:
    from collections import OrderedDict as _DictReadMapping


if TYPE_CHECKING:
    from _typeshed import FileDescriptorOrPath  # NOQA
    from _typeshed import SupportsIter  # NOQA
    from _typeshed import SupportsWrite  # NOQA

    __all__ = [
        "SupportsWrite",
        "SupportsIter",
        "FileDescriptorOrPath",
        "AnyPath",
        "_OpenFile",
        "_DictRow",
        "_DialectLike",
        "_DictReadMapping",
    ]
else:
    __all__ = [
        "AnyPath",
        "_OpenFile",
        "_DictRow",
        "_DialectLike",
        "_DictReadMapping",
    ]



import io

from typing import Any

basestring = str
file = io.FileIO

def KEY_CTRL(key): ...
def addstr(*args): ...
def insstr(*args): ...

class ReloadException(Exception):
    start_pos: Any
    column_width_mode: Any
    column_gap: Any
    column_widths: Any
    search_str: Any
    def __init__(
        self, start_pos, column_width, column_gap, column_widths, search_str
    ) -> None: ...

class QuitException(Exception): ...

class Viewer:
    scr: Any
    data: Any
    info: Any
    header_offset_orig: int
    header: Any
    header_offset: Any
    num_data_columns: Any
    column_width_mode: Any
    column_gap: Any
    trunc_char: Any
    num_columns: int
    vis_columns: int
    init_search: Any
    modifier: Any
    def __init__(self, *args, **kwargs) -> None: ...
    def column_xw(self, x): ...
    def quit(self) -> None: ...
    def reload(self) -> None: ...
    def consume_modifier(self, default: int = ...): ...
    def down(self) -> None: ...
    def up(self) -> None: ...
    def left(self) -> None: ...
    def right(self) -> None: ...
    y: Any
    win_y: Any
    def page_down(self) -> None: ...
    def page_up(self) -> None: ...
    x: Any
    win_x: Any
    def page_right(self) -> None: ...
    def page_left(self) -> None: ...
    def mark(self) -> None: ...
    def goto_mark(self) -> None: ...
    def home(self) -> None: ...
    def goto_y(self, y) -> None: ...
    def goto_row(self) -> None: ...
    def goto_x(self, x) -> None: ...
    def goto_col(self) -> None: ...
    def goto_yx(self, y, x) -> None: ...
    def line_home(self) -> None: ...
    def line_end(self) -> None: ...
    def show_cell(self) -> None: ...
    def show_info(self): ...
    textpad: Any
    search_str: Any
    def search(self) -> None: ...
    def search_results(
        self, rev: bool = ..., look_in_cur: bool = ...
    ) -> None: ...
    def search_results_prev(
        self, rev: bool = ..., look_in_cur: bool = ...
    ) -> None: ...
    def help(self) -> None: ...
    def toggle_header(self) -> None: ...
    def column_gap_down(self) -> None: ...
    def column_gap_up(self) -> None: ...
    column_width: Any
    def column_width_all_down(self) -> None: ...
    def column_width_all_up(self) -> None: ...
    def column_width_down(self) -> None: ...
    def column_width_up(self) -> None: ...
    def sort_by_column_numeric(self): ...
    def sort_by_column_numeric_reverse(self): ...
    def sort_by_column(self) -> None: ...
    def sort_by_column_reverse(self) -> None: ...
    def sort_by_column_natural(self) -> None: ...
    def sort_by_column_natural_reverse(self) -> None: ...
    def sorted_nicely(self, ls, key, rev: bool = ...): ...
    def float_string_key(self, value): ...
    def toggle_column_width(self) -> None: ...
    def set_current_column_width(self) -> None: ...
    def yank_cell(self) -> None: ...
    keys: Any
    def define_keys(self) -> None: ...
    def run(self) -> None: ...
    def handle_keys(self) -> None: ...
    def handle_modifier(self, mod) -> None: ...
    def resize(self) -> None: ...
    def num_columns_fwd(self, x): ...
    def num_columns_rev(self, x): ...
    def recalculate_layout(self) -> None: ...
    def location_string(self, yp, xp): ...
    def display(self) -> None: ...
    def strpad(self, s, width): ...
    def hdrstr(self, x, width): ...
    def cellstr(self, y, x, width): ...
    def skip_to_row_change(self) -> None: ...
    def skip_to_row_change_reverse(self) -> None: ...
    def skip_to_col_change(self) -> None: ...
    def skip_to_col_change_reverse(self) -> None: ...

class TextBox:
    scr: Any
    data: Any
    title: Any
    tdata: Any
    hid_rows: int
    def __init__(self, scr, data: str = ..., title: str = ...) -> None: ...
    def __call__(self) -> None: ...
    handlers: Any
    def setup_handlers(self) -> None: ...
    def run(self) -> None: ...
    def handle_key(self, key) -> None: ...
    def close(self) -> None: ...
    def scroll_down(self) -> None: ...
    def scroll_up(self) -> None: ...
    def display(self) -> None: ...

def csv_sniff(data, enc): ...
def fix_newlines(data): ...
def adjust_space_delim(data, enc): ...
def process_data(
    data,
    enc: Any | None = ...,
    delim: Any | None = ...,
    quoting: Any | None = ...,
    quote_char=...,
): ...
def data_list_or_file(data): ...
def pad_data(d): ...
def readme(): ...
def detect_encoding(data: Any | None = ...): ...
def main(stdscr, *args, **kwargs) -> None: ...
def view(
    data,
    enc: Any | None = ...,
    start_pos=...,
    column_width: int = ...,
    column_gap: int = ...,
    trunc_char: str = ...,
    column_widths: Any | None = ...,
    search_str: Any | None = ...,
    double_width: bool = ...,
    delimiter: Any | None = ...,
    quoting: Any | None = ...,
    info: Any | None = ...,
    quote_char=...,
): ...
def parse_path(path): ...



# -*- coding: utf-8 -*-

from typing import Any
from typing import Iterable
from typing import Iterator
from typing import List
from typing import Literal
from typing import Optional
from typing import Tuple
from typing import Union
from typing import overload

from .dialect import SimpleDialect

def field_size_limit(*args: Any, **kwargs: Any) -> int: ...
@overload
def _parse_data(
    data: Iterable[str],
    delimiter: str,
    quotechar: str,
    escapechar: str,
    strict: bool,
    return_quoted: Literal[False] = ...,
) -> Iterator[List[str]]: ...
@overload
def _parse_data(
    data: Iterable[str],
    delimiter: str,
    quotechar: str,
    escapechar: str,
    strict: bool,
    return_quoted: Literal[True],
) -> Iterator[List[Tuple[str, bool]]]: ...
@overload
def _parse_data(
    data: Iterable[str],
    delimiter: str,
    quotechar: str,
    escapechar: str,
    strict: bool,
    return_quoted: bool = ...,
) -> Iterator[Union[List[str], List[Tuple[str, bool]]]]: ...
def parse_data(
    data: Iterable[str],
    dialect: Optional[SimpleDialect] = None,
    delimiter: Optional[str] = None,
    quotechar: Optional[str] = None,
    escapechar: Optional[str] = None,
    strict: Optional[bool] = None,
    return_quoted: bool = False,
) -> Iterator[Union[List[str], List[Tuple[str, bool]]]]: ...
@overload
def parse_string(
    data: str,
    dialect: SimpleDialect,
    return_quoted: Literal[False] = ...,
) -> Iterator[List[str]]: ...
@overload
def parse_string(
    data: str,
    dialect: SimpleDialect,
    return_quoted: Literal[True],
) -> Iterator[List[Tuple[str, bool]]]: ...
@overload
def parse_string(
    data: str,
    dialect: SimpleDialect,
    return_quoted: bool = ...,
) -> Iterator[Union[List[str], List[Tuple[str, bool]]]]: ...



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Generate strings for all the date formats.

Author: Gertjan van den Burg

"""


def main():
    year2 = "YY"
    year4 = "YYYY"
    month_leading = "MM"
    month_sparse = "M"
    day_leading = "D"
    day_leading = "DD"
    day_sparse = "D"
    sep = "x"
    pats = []
    for year in [year2, year4]:
        for month in [month_leading, month_sparse]:
            for day in [day_leading, day_sparse]:
                fmt = dict(year=year, month=month, day=day, sep=sep)
                pats.append("{year}{sep}{month}{sep}{day}".format(**fmt))
                pats.append("{day}{sep}{month}{sep}{year}".format(**fmt))
                pats.append("{month}{sep}{day}{sep}{year}".format(**fmt))
                pats.append("{year}年{month}月{day}日".format(**fmt))
                pats.append("{year}년{month}월{day}일".format(**fmt))

    for year in [year2, year4]:
        fmt = dict(year=year, month=month_leading, day=day_leading, sep="")
        pats.append("{year}{sep}{month}{sep}{day}".format(**fmt))
        pats.append("{day}{sep}{month}{sep}{year}".format(**fmt))
        pats.append("{month}{sep}{day}{sep}{year}".format(**fmt))

    for pat in sorted(pats):
        print(pat)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

"""
Unit tests for the CSV parser.

Author: Gertjan van den Burg

"""

import io
import unittest

from typing import Any
from typing import List
from typing import Tuple
from typing import TypeVar

from clevercsv.cparser_util import parse_data

T = TypeVar("T", str, Tuple[str, bool])


class ParserTestCase(unittest.TestCase):

    """
    Testing splitting on delimiter with or without quotes
    """

    def _parse_test(
        self, string: str, expect: List[List[T]], **kwargs: Any
    ) -> None:
        buf = io.StringIO(string, newline="")
        result = list(parse_data(buf, **kwargs))
        self.assertEqual(result, expect)

    def test_parse_simple_1(self) -> None:
        self._parse_test(
            "A,B,C,D,E",
            [["A", "B", "C", "D", "E"]],
            delimiter=",",
            quotechar='"',
        )

    def test_parse_simple_2(self) -> None:
        self._parse_test(
            "A,B,C,D,E",
            [["A", "B", "C", "D", "E"]],
            delimiter=",",
            quotechar="",
        )

    def test_parse_simple_3(self) -> None:
        self._parse_test("A,B,C,D,E", [["A,B,C,D,E"]])

    def test_parse_simple_4(self) -> None:
        self._parse_test(
            'A,"B",C,D,E',
            [["A", "B", "C", "D", "E"]],
            delimiter=",",
            quotechar='"',
        )

    def test_parse_simple_5(self) -> None:
        self._parse_test(
            'A,"B,C",D,E',
            [["A", "B,C", "D", "E"]],
            delimiter=",",
            quotechar='"',
        )

    def test_parse_simple_6(self) -> None:
        self._parse_test(
            'A,"B,C",D,E',
            [["A", '"B', 'C"', "D", "E"]],
            delimiter=",",
            quotechar="",
        )

    def test_parse_simple_7(self) -> None:
        self._parse_test(
            '"A","B","C",,,,',
            [['"A"', '"B"', '"C"', "", "", "", ""]],
            delimiter=",",
            quotechar="",
        )

    """
    Testing splitting on rows only:
    """

    def test_parse_no_delim_1(self) -> None:
        self._parse_test(
            'A"B"C\rA"B""C""D"', [['A"B"C'], ['A"B""C""D"']], quotechar=""
        )

    def test_parse_no_delim_2(self) -> None:
        self._parse_test(
            'A"B"C\nA"B""C""D"', [['A"B"C'], ['A"B""C""D"']], quotechar=""
        )

    def test_parse_no_delim_3(self) -> None:
        self._parse_test(
            'A"B"C\r\nA"B""C""D"', [['A"B"C'], ['A"B""C""D"']], quotechar=""
        )

    def test_parse_no_delim_4(self) -> None:
        self._parse_test(
            'A"B\r\nB"C\r\nD"E"F\r\nG',
            [['A"B\r\nB"C'], ['D"E"F'], ["G"]],
            quotechar='"',
        )

    def test_parse_no_delim_5(self) -> None:
        self._parse_test(
            'A"B\nB"C\nD"E"F\nG',
            [['A"B\nB"C'], ['D"E"F'], ["G"]],
            quotechar='"',
        )

    def test_parse_no_delim_6(self) -> None:
        self._parse_test(
            'A"B\nB\rB"C\nD"E"F\nG',
            [['A"B\nB\rB"C'], ['D"E"F'], ["G"]],
            quotechar='"',
        )

    """
    Tests from Pythons builtin CSV module:
    """

    def test_parse_builtin_1(self) -> None:
        self._parse_test("", [])

    def test_parse_builtin_2(self) -> None:
        self._parse_test("a,b\r", [["a", "b"]], delimiter=",")

    def test_parse_builtin_3(self) -> None:
        self._parse_test("a,b\n", [["a", "b"]], delimiter=",")

    def test_parse_builtin_4(self) -> None:
        self._parse_test("a,b\r\n", [["a", "b"]], delimiter=",")

    def test_parse_builtin_5(self) -> None:
        self._parse_test('a,"', [["a", ""]], delimiter=",", quotechar='"')

    def test_parse_builtin_6(self) -> None:
        self._parse_test('"a', [["a"]], delimiter=",", quotechar='"')

    def test_parse_builtin_7(self) -> None:
        # differs from Python (1)
        self._parse_test(
            "a,|b,c",
            [["a", "|b", "c"]],
            delimiter=",",
            quotechar='"',
            escapechar="|",
        )

    def test_parse_builtin_8(self) -> None:
        self._parse_test(
            "a,b|,c",
            [["a", "b,c"]],
            delimiter=",",
            quotechar='"',
            escapechar="|",
        )

    def test_parse_builtin_9(self) -> None:
        # differs from Python (1)
        self._parse_test(
            'a,"b,|c"',
            [["a", "b,|c"]],
            delimiter=",",
            quotechar='"',
            escapechar="|",
        )

    def test_parse_builtin_10(self) -> None:
        self._parse_test(
            'a,"b,c|""',
            [["a", 'b,c"']],
            delimiter=",",
            quotechar='"',
            escapechar="|",
        )

    def test_parse_builtin_11(self) -> None:
        # differs from Python (2)
        self._parse_test(
            'a,"b,c"|',
            [["a", "b,c"]],
            delimiter=",",
            quotechar='"',
            escapechar="|",
        )

    def test_parse_builtin_12(self) -> None:
        self._parse_test(
            '1,",3,",5', [["1", ",3,", "5"]], delimiter=",", quotechar='"'
        )

    def test_parse_builtin_13(self) -> None:
        self._parse_test(
            '1,",3,",5',
            [["1", '"', "3", '"', "5"]],
            delimiter=",",
            quotechar="",
        )

    def test_parse_builtin_14(self) -> None:
        self._parse_test(
            ',3,"5",7.3, 9',
            [["", "3", "5", "7.3", " 9"]],
            delimiter=",",
            quotechar='"',
        )

    def test_parse_builtin_15(self) -> None:
        self._parse_test(
            '"a\nb", 7', [["a\nb", " 7"]], delimiter=",", quotechar='"'
        )

    """
    Double quotes:
    """

    def test_parse_dq_1(self) -> None:
        self._parse_test(
            'a,"a""b""c"', [["a", 'a"b"c']], delimiter=",", quotechar='"'
        )

    def test_parse_dq_2(self) -> None:
        self._parse_test(
            'a,"a""b,c""d",e',
            [["a", 'a"b,c"d', "e"]],
            delimiter=",",
            quotechar='"',
        )

    """
    Mix double and escapechar:
    """

    def test_parse_mix_double_escape_1(self) -> None:
        self._parse_test(
            'a,"bc""d"",|"f|""',
            [["a", 'bc"d","f"']],
            delimiter=",",
            quotechar='"',
            escapechar="|",
        )

    """
    Other tests:
    """

    def test_parse_other_1(self) -> None:
        self._parse_test(
            'a,b "c" d,e', [["a", 'b "c" d', "e"]], delimiter=",", quotechar=""
        )

    def test_parse_other_2(self) -> None:
        self._parse_test(
            'a,b "c" d,e',
            [["a", 'b "c" d', "e"]],
            delimiter=",",
            quotechar='"',
        )

    def test_parse_other_3(self) -> None:
        self._parse_test("a,\rb,c", [["a", ""], ["b", "c"]], delimiter=",")

    def test_parse_other_4(self) -> None:
        self._parse_test(
            "a,b\r\n\r\nc,d\r\n", [["a", "b"], [], ["c", "d"]], delimiter=","
        )

    def test_parse_other_5(self) -> None:
        self._parse_test(
            "\r\na,b\rc,d\n\re,f\r\n",
            [[], ["a", "b"], ["c", "d"], [], ["e", "f"]],
            delimiter=",",
        )

    def test_parse_other_6(self) -> None:
        self._parse_test(
            "a,b\n\nc,d", [["a", "b"], [], ["c", "d"]], delimiter=","
        )

    """
    Further escape char tests:
    """

    def test_parse_escape_1(self) -> None:
        self._parse_test(
            "a,b,c||d",
            [["a", "b", "c|d"]],
            delimiter=",",
            quotechar="",
            escapechar="|",
        )

    def test_parse_escape_2(self) -> None:
        self._parse_test(
            "a,b,c||d,e|,d",
            [["a", "b", "c|d", "e,d"]],
            delimiter=",",
            quotechar="",
            escapechar="|",
        )

    """
    Quote mismatch until EOF:
    """

    def test_parse_quote_mismatch_1(self) -> None:
        self._parse_test(
            'a,b,c"d,e\n',
            [["a", "b", 'c"d,e\n']],
            delimiter=",",
            quotechar='"',
        )

    def test_parse_quote_mismatch_2(self) -> None:
        self._parse_test(
            'a,b,c"d,e\n',
            [["a", "b", 'c"d', "e"]],
            delimiter=",",
            quotechar="",
        )

    def test_parse_quote_mismatch_3(self) -> None:
        self._parse_test(
            'a,b,"c,d', [["a", "b", "c,d"]], delimiter=",", quotechar='"'
        )

    def test_parse_quote_mismatch_4(self) -> None:
        self._parse_test(
            'a,b,"c,d\n', [["a", "b", "c,d\n"]], delimiter=",", quotechar='"'
        )

    """
    Single column:
    """

    def test_parse_single_1(self) -> None:
        self._parse_test("a\rb\rc\n", [["a"], ["b"], ["c"]])

    """
    These tests illustrate a difference with the Python parser, which in this
    case would return ``[['a', 'abc', 'd']]``.
    """

    def test_parse_differ_1(self) -> None:
        self._parse_test(
            'a,"ab"c,d', [["a", '"ab"c', "d"]], delimiter=",", quotechar=""
        )

    def test_parse_differ_2(self) -> None:
        self._parse_test(
            'a,"ab"c,d', [["a", '"ab"c', "d"]], delimiter=",", quotechar='"'
        )

    """
    Return quoted
    """

    def test_parse_return_quoted_1(self) -> None:
        self._parse_test(
            "a,b,c",
            [[("a", False), ("b", False), ("c", False)]],
            delimiter=",",
            quotechar="",
            return_quoted=True,
        )

    def test_parse_return_quoted_2(self) -> None:
        self._parse_test(
            'a,"b,c",d',
            [[("a", False), ("b,c", True), ("d", False)]],
            delimiter=",",
            quotechar='"',
            return_quoted=True,
        )

    def test_parse_return_quoted_3(self) -> None:
        self._parse_test(
            'a,b,"c,d',
            [[("a", False), ("b", False), ("c,d", True)]],
            delimiter=",",
            quotechar='"',
            return_quoted=True,
        )


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-

"""
Unit tests for the potential dialect selection.

Author: Gertjan van den Burg

"""

import unittest

from clevercsv.potential_dialects import filter_urls
from clevercsv.potential_dialects import get_delimiters
from clevercsv.potential_dialects import get_quotechars
from clevercsv.potential_dialects import masked_by_quotechar


class PotentialDialectTestCase(unittest.TestCase):
    def test_masked_by_quotechar(self) -> None:
        self.assertTrue(masked_by_quotechar('A"B&C"A', '"', "", "&"))
        self.assertFalse(masked_by_quotechar('A"B&C"A&A', '"', "", "&"))
        self.assertFalse(masked_by_quotechar('A|"B&C"A', '"', "|", "&"))
        self.assertFalse(masked_by_quotechar('A"B"C', '"', "", ""))

    def test_filter_urls(self) -> None:
        data = "A,B\nwww.google.com,10\nhttps://gertjanvandenburg.com,25\n"
        exp = "A,B\nU,10\nU,25\n"
        self.assertEqual(exp, filter_urls(data))

    def test_get_quotechars(self) -> None:
        data = "A,B,'A',B\"D\"E"
        exp = set(['"', "'", ""])
        out = get_quotechars(data)
        self.assertEqual(out, exp)

    def test_get_delimiters(self) -> None:
        data = "A,B|CD,E;F\tD123£123€10.,0"
        exp = set([",", "|", ";", "\t", "€", "£", ""])
        out = get_delimiters(data, "UTF-8")
        self.assertEqual(out, exp)


if __name__ == "__main__":
    unittest.main()



from typing import Any as _Any

class error(Exception):
    msg: _Any
    pattern: _Any
    pos: _Any
    lineno: _Any
    colno: _Any
    def __init__(
        self, message, pattern: _Any | None = ..., pos: _Any | None = ...
    ) -> None: ...

class _UnscopedFlagSet(Exception): ...
class ParseError(Exception): ...
class _FirstSetError(Exception): ...

A: int

ASCII: int
B: int
BESTMATCH: int
D: int
DEBUG: int
E: int
ENHANCEMATCH: int
F: int
FULLCASE: int
I: int
IGNORECASE: int
L: int
LOCALE: int
M: int
MULTILINE: int
P: int
POSIX: int
R: int
REVERSE: int
S: int
DOTALL: int
U: int
UNICODE: int
V0: int
VERSION0: int
V1: int
VERSION1: int
W: int
WORD: int
X: int
VERBOSE: int
T: int
TEMPLATE: int
DEFAULT_VERSION = VERSION1

class Namespace: ...

class RegexBase:
    def __init__(self) -> None: ...
    def with_flags(
        self,
        positive: _Any | None = ...,
        case_flags: _Any | None = ...,
        zerowidth: _Any | None = ...,
    ): ...
    def fix_groups(self, pattern, reverse, fuzzy) -> None: ...
    def optimise(self, info, reverse): ...
    def pack_characters(self, info): ...
    def remove_captures(self): ...
    def is_atomic(self): ...
    def can_be_affix(self): ...
    def contains_group(self): ...
    def get_firstset(self, reverse) -> None: ...
    def has_simple_start(self): ...
    def compile(self, reverse: bool = ..., fuzzy: bool = ...): ...
    def is_empty(self): ...
    def __hash__(self): ...
    def __eq__(self, other): ...
    def __ne__(self, other): ...
    def get_required_string(self, reverse): ...

class ZeroWidthBase(RegexBase):
    positive: _Any
    def __init__(self, positive: bool = ...) -> None: ...
    def get_firstset(self, reverse): ...
    def dump(self, indent, reverse) -> None: ...
    def max_width(self): ...

class Any(RegexBase):
    def has_simple_start(self): ...
    def dump(self, indent, reverse) -> None: ...
    def max_width(self): ...

class AnyAll(Any): ...
class AnyU(Any): ...

class Atomic(RegexBase):
    subpattern: _Any
    def __init__(self, subpattern) -> None: ...
    def fix_groups(self, pattern, reverse, fuzzy) -> None: ...
    def optimise(self, info, reverse): ...
    def pack_characters(self, info): ...
    def remove_captures(self): ...
    def can_be_affix(self): ...
    def contains_group(self): ...
    def get_firstset(self, reverse): ...
    def has_simple_start(self): ...
    def dump(self, indent, reverse) -> None: ...
    def is_empty(self): ...
    def __eq__(self, other): ...
    def max_width(self): ...
    def get_required_string(self, reverse): ...

class Boundary(ZeroWidthBase): ...

class Branch(RegexBase):
    branches: _Any
    def __init__(self, branches) -> None: ...
    def fix_groups(self, pattern, reverse, fuzzy) -> None: ...
    def optimise(self, info, reverse): ...
    def pack_characters(self, info): ...
    def remove_captures(self): ...
    def is_atomic(self): ...
    def can_be_affix(self): ...
    def contains_group(self): ...
    def get_firstset(self, reverse): ...
    def dump(self, indent, reverse) -> None: ...
    def is_empty(self): ...
    def __eq__(self, other): ...
    def max_width(self): ...

class CallGroup(RegexBase):
    info: _Any
    group: _Any
    position: _Any
    def __init__(self, info, group, position) -> None: ...
    def fix_groups(self, pattern, reverse, fuzzy) -> None: ...
    def remove_captures(self) -> None: ...
    def dump(self, indent, reverse) -> None: ...
    def __eq__(self, other): ...
    def max_width(self): ...
    def __del__(self) -> None: ...

class CallRef(RegexBase):
    ref: _Any
    parsed: _Any
    def __init__(self, ref, parsed) -> None: ...

class Character(RegexBase):
    value: _Any
    positive: _Any
    case_flags: _Any
    zerowidth: _Any
    folded: _Any
    def __init__(
        self,
        value,
        positive: bool = ...,
        case_flags=...,
        zerowidth: bool = ...,
    ) -> None: ...
    def rebuild(self, positive, case_flags, zerowidth): ...
    def optimise(self, info, reverse, in_set: bool = ...): ...
    def get_firstset(self, reverse): ...
    def has_simple_start(self): ...
    def dump(self, indent, reverse) -> None: ...
    def matches(self, ch): ...
    def max_width(self): ...
    folded_characters: _Any
    def get_required_string(self, reverse): ...

class Conditional(RegexBase):
    info: _Any
    group: _Any
    yes_item: _Any
    no_item: _Any
    position: _Any
    def __init__(self, info, group, yes_item, no_item, position) -> None: ...
    def fix_groups(self, pattern, reverse, fuzzy) -> None: ...
    def optimise(self, info, reverse): ...
    def pack_characters(self, info): ...
    def remove_captures(self) -> None: ...
    def is_atomic(self): ...
    def can_be_affix(self): ...
    def contains_group(self): ...
    def get_firstset(self, reverse): ...
    def dump(self, indent, reverse) -> None: ...
    def is_empty(self): ...
    def __eq__(self, other): ...
    def max_width(self): ...
    def __del__(self) -> None: ...

class DefaultBoundary(ZeroWidthBase): ...
class DefaultEndOfWord(ZeroWidthBase): ...
class DefaultStartOfWord(ZeroWidthBase): ...
class EndOfLine(ZeroWidthBase): ...
class EndOfLineU(EndOfLine): ...
class EndOfString(ZeroWidthBase): ...
class EndOfStringLine(ZeroWidthBase): ...
class EndOfStringLineU(EndOfStringLine): ...
class EndOfWord(ZeroWidthBase): ...
class Failure(ZeroWidthBase): ...

class Fuzzy(RegexBase):
    subpattern: _Any
    constraints: _Any
    def __init__(self, subpattern, constraints: _Any | None = ...) -> None: ...
    def fix_groups(self, pattern, reverse, fuzzy) -> None: ...
    def pack_characters(self, info): ...
    def remove_captures(self): ...
    def is_atomic(self): ...
    def contains_group(self): ...
    def dump(self, indent, reverse) -> None: ...
    def is_empty(self): ...
    def __eq__(self, other): ...
    def max_width(self): ...

class Grapheme(RegexBase):
    def dump(self, indent, reverse) -> None: ...
    def max_width(self): ...

class GraphemeBoundary:
    def compile(self, reverse, fuzzy): ...

class GreedyRepeat(RegexBase):
    subpattern: _Any
    min_count: _Any
    max_count: _Any
    def __init__(self, subpattern, min_count, max_count) -> None: ...
    def fix_groups(self, pattern, reverse, fuzzy) -> None: ...
    def optimise(self, info, reverse): ...
    def pack_characters(self, info): ...
    def remove_captures(self): ...
    def is_atomic(self): ...
    def can_be_affix(self): ...
    def contains_group(self): ...
    def get_firstset(self, reverse): ...
    def dump(self, indent, reverse) -> None: ...
    def is_empty(self): ...
    def __eq__(self, other): ...
    def max_width(self): ...
    def get_required_string(self, reverse): ...

class PossessiveRepeat(GreedyRepeat):
    def is_atomic(self): ...
    def dump(self, indent, reverse) -> None: ...

class Group(RegexBase):
    info: _Any
    group: _Any
    subpattern: _Any
    call_ref: _Any
    def __init__(self, info, group, subpattern) -> None: ...
    def fix_groups(self, pattern, reverse, fuzzy) -> None: ...
    def optimise(self, info, reverse): ...
    def pack_characters(self, info): ...
    def remove_captures(self): ...
    def is_atomic(self): ...
    def can_be_affix(self): ...
    def contains_group(self): ...
    def get_firstset(self, reverse): ...
    def has_simple_start(self): ...
    def dump(self, indent, reverse) -> None: ...
    def __eq__(self, other): ...
    def max_width(self): ...
    def get_required_string(self, reverse): ...
    def __del__(self) -> None: ...

class Keep(ZeroWidthBase): ...
class LazyRepeat(GreedyRepeat): ...

class LookAround(RegexBase):
    behind: _Any
    positive: _Any
    subpattern: _Any
    def __init__(self, behind, positive, subpattern) -> None: ...
    def fix_groups(self, pattern, reverse, fuzzy) -> None: ...
    def optimise(self, info, reverse): ...
    def pack_characters(self, info): ...
    def remove_captures(self): ...
    def is_atomic(self): ...
    def can_be_affix(self): ...
    def contains_group(self): ...
    def get_firstset(self, reverse): ...
    def dump(self, indent, reverse) -> None: ...
    def is_empty(self): ...
    def __eq__(self, other): ...
    def max_width(self): ...

class LookAroundConditional(RegexBase):
    behind: _Any
    positive: _Any
    subpattern: _Any
    yes_item: _Any
    no_item: _Any
    def __init__(
        self, behind, positive, subpattern, yes_item, no_item
    ) -> None: ...
    def fix_groups(self, pattern, reverse, fuzzy) -> None: ...
    def optimise(self, info, reverse): ...
    def pack_characters(self, info): ...
    def remove_captures(self) -> None: ...
    def is_atomic(self): ...
    def can_be_affix(self): ...
    def contains_group(self): ...
    def dump(self, indent, reverse) -> None: ...
    def is_empty(self): ...
    def __eq__(self, other): ...
    def max_width(self): ...
    def get_required_string(self, reverse): ...

class PrecompiledCode(RegexBase):
    code: _Any
    def __init__(self, code) -> None: ...

class Property(RegexBase):
    value: _Any
    positive: _Any
    case_flags: _Any
    zerowidth: _Any
    def __init__(
        self,
        value,
        positive: bool = ...,
        case_flags=...,
        zerowidth: bool = ...,
    ) -> None: ...
    def rebuild(self, positive, case_flags, zerowidth): ...
    def optimise(self, info, reverse, in_set: bool = ...): ...
    def get_firstset(self, reverse): ...
    def has_simple_start(self): ...
    def dump(self, indent, reverse) -> None: ...
    def matches(self, ch): ...
    def max_width(self): ...

class Prune(ZeroWidthBase): ...

class Range(RegexBase):
    lower: _Any
    upper: _Any
    positive: _Any
    case_flags: _Any
    zerowidth: _Any
    def __init__(
        self,
        lower,
        upper,
        positive: bool = ...,
        case_flags=...,
        zerowidth: bool = ...,
    ) -> None: ...
    def rebuild(self, positive, case_flags, zerowidth): ...
    def optimise(self, info, reverse, in_set: bool = ...): ...
    def dump(self, indent, reverse) -> None: ...
    def matches(self, ch): ...
    def max_width(self): ...

class RefGroup(RegexBase):
    info: _Any
    group: _Any
    position: _Any
    case_flags: _Any
    def __init__(self, info, group, position, case_flags=...) -> None: ...
    def fix_groups(self, pattern, reverse, fuzzy) -> None: ...
    def remove_captures(self) -> None: ...
    def dump(self, indent, reverse) -> None: ...
    def max_width(self): ...
    def __del__(self) -> None: ...

class SearchAnchor(ZeroWidthBase): ...

class Sequence(RegexBase):
    items: _Any
    def __init__(self, items: _Any | None = ...) -> None: ...
    def fix_groups(self, pattern, reverse, fuzzy) -> None: ...
    def optimise(self, info, reverse): ...
    def pack_characters(self, info): ...
    def remove_captures(self): ...
    def is_atomic(self): ...
    def can_be_affix(self): ...
    def contains_group(self): ...
    def get_firstset(self, reverse): ...
    def has_simple_start(self): ...
    def dump(self, indent, reverse) -> None: ...
    def is_empty(self): ...
    def __eq__(self, other): ...
    def max_width(self): ...
    def get_required_string(self, reverse): ...

class SetBase(RegexBase):
    info: _Any
    items: _Any
    positive: _Any
    case_flags: _Any
    zerowidth: _Any
    char_width: int
    def __init__(
        self,
        info,
        items,
        positive: bool = ...,
        case_flags=...,
        zerowidth: bool = ...,
    ) -> None: ...
    def rebuild(self, positive, case_flags, zerowidth): ...
    def get_firstset(self, reverse): ...
    def has_simple_start(self): ...
    def dump(self, indent, reverse) -> None: ...
    def max_width(self): ...
    def __del__(self) -> None: ...

class SetDiff(SetBase):
    items: _Any
    def optimise(self, info, reverse, in_set: bool = ...): ...
    def matches(self, ch): ...

class SetInter(SetBase):
    items: _Any
    def optimise(self, info, reverse, in_set: bool = ...): ...
    def matches(self, ch): ...

class SetSymDiff(SetBase):
    items: _Any
    def optimise(self, info, reverse, in_set: bool = ...): ...
    def matches(self, ch): ...

class SetUnion(SetBase):
    items: _Any
    def optimise(self, info, reverse, in_set: bool = ...): ...
    def matches(self, ch): ...

class Skip(ZeroWidthBase): ...
class StartOfLine(ZeroWidthBase): ...
class StartOfLineU(StartOfLine): ...
class StartOfString(ZeroWidthBase): ...
class StartOfWord(ZeroWidthBase): ...

class String(RegexBase):
    characters: _Any
    case_flags: _Any
    folded_characters: _Any
    required: bool
    def __init__(self, characters, case_flags=...) -> None: ...
    def get_firstset(self, reverse): ...
    def has_simple_start(self): ...
    def dump(self, indent, reverse) -> None: ...
    def max_width(self): ...
    def get_required_string(self, reverse): ...

class Literal(String):
    def dump(self, indent, reverse) -> None: ...

class StringSet(Branch):
    info: _Any
    name: _Any
    case_flags: _Any
    set_key: _Any
    branches: _Any
    def __init__(self, info, name, case_flags=...) -> None: ...
    def dump(self, indent, reverse) -> None: ...
    def __del__(self) -> None: ...

class Source:
    string: _Any
    char_type: _Any
    pos: int
    ignore_space: bool
    sep: _Any
    def __init__(self, string): ...
    def get(self, override_ignore: bool = ...): ...
    def get_many(self, count: int = ...): ...
    def get_while(self, test_set, include: bool = ...): ...
    def skip_while(self, test_set, include: bool = ...) -> None: ...
    def match(self, substring): ...
    def expect(self, substring) -> None: ...
    def at_end(self): ...

class Info:
    flags: _Any
    global_flags: _Any
    inline_locale: bool
    kwargs: _Any
    group_count: int
    group_index: _Any
    group_name: _Any
    char_type: _Any
    named_lists_used: _Any
    open_groups: _Any
    open_group_count: _Any
    defined_groups: _Any
    group_calls: _Any
    private_groups: _Any
    def __init__(
        self, flags: int = ..., char_type: _Any | None = ..., kwargs=...
    ) -> None: ...
    def open_group(self, name: _Any | None = ...): ...
    def close_group(self) -> None: ...
    def is_open_group(self, name): ...

class Scanner:
    lexicon: _Any
    scanner: _Any
    def __init__(self, lexicon, flags: int = ...) -> None: ...
    match: _Any
    def scan(self, string): ...



# -*- coding: utf-8 -*-

"""
Drop-in replacement for the Python csv reader class. This is a wrapper for the
Parser class, defined in :mod:`cparser`.

Author: Gertjan van den Burg

"""

import csv

from typing import Any
from typing import Iterable
from typing import Iterator
from typing import List
from typing import Optional

from . import field_size_limit
from ._types import _DialectLike
from .cparser import Error as ParserError
from .cparser import Parser
from .dialect import SimpleDialect
from .exceptions import Error


class reader:
    def __init__(
        self,
        csvfile: Iterable[str],
        dialect: _DialectLike = "excel",
        **fmtparams: Any,
    ):
        self.csvfile = csvfile
        self.original_dialect = dialect
        self._dialect = self._make_simple_dialect(dialect, **fmtparams)
        self.line_num: int = 0
        self.parser_gen: Optional[Parser] = None

    @property
    def dialect(self) -> csv.Dialect:
        return self._dialect.to_csv_dialect()

    def _make_simple_dialect(
        self, dialect: _DialectLike, **fmtparams: Any
    ) -> SimpleDialect:
        if isinstance(dialect, str):
            sd = SimpleDialect.from_csv_dialect(csv.get_dialect(dialect))
        elif isinstance(dialect, csv.Dialect):
            sd = SimpleDialect.from_csv_dialect(dialect)
        elif isinstance(dialect, SimpleDialect):
            sd = dialect
        else:
            raise ValueError("Unknown dialect type: %r" % dialect)
        for key, value in fmtparams.items():
            if key in ["delimiter", "quotechar", "escapechar", "strict"]:
                setattr(sd, key, value)
        sd.validate()
        return sd

    def __iter__(self) -> Iterator[List[str]]:
        self.parser_gen = Parser(
            self.csvfile,
            delimiter=self._dialect.delimiter,
            quotechar=self._dialect.quotechar,
            escapechar=self._dialect.escapechar,
            field_limit=field_size_limit(),
            strict=self._dialect.strict,
        )
        return self

    def __next__(self) -> List[str]:
        if self.parser_gen is None:
            self.__iter__()
        assert self.parser_gen is not None
        try:
            row = next(self.parser_gen)
        except ParserError as e:
            raise Error(str(e))
        self.line_num += 1
        return row



# -*- coding: utf-8 -*-

"""
Unit tests for the CCSV write module.

Author: Gertjan van den Burg

"""


import csv
import tempfile
import unittest

from typing import Any
from typing import Iterable
from typing import Type

import clevercsv

from clevercsv.dialect import SimpleDialect


class WriterTestCase(unittest.TestCase):
    def _write_test(
        self, fields: Iterable[Any], expect: str, **kwargs: Any
    ) -> None:
        with tempfile.TemporaryFile("w+", newline="", prefix="ccsv_") as fp:
            writer = clevercsv.writer(fp, **kwargs)
            writer.writerow(fields)
            fp.seek(0)
            self.assertEqual(fp.read(), expect + writer.dialect.lineterminator)

    def _write_error_test(
        self, exc: Type[Exception], fields: Any, **kwargs: Any
    ) -> None:
        with tempfile.TemporaryFile("w+", newline="", prefix="ccsv_") as fp:
            writer = clevercsv.writer(fp, **kwargs)
            with self.assertRaises(exc):
                writer.writerow(fields)
            fp.seek(0)
            self.assertEqual(fp.read(), "")

    def test_write_arg_valid(self) -> None:
        self._write_error_test(clevercsv.Error, None)
        self._write_test((), "")
        self._write_test([None], '""')
        self._write_error_test(
            clevercsv.Error, [None], quoting=clevercsv.QUOTE_NONE
        )

        # Check that exceptions are passed up the chain
        class BadList:
            def __len__(self) -> int:
                return 10

            def __getitem__(self, i: int) -> None:
                if i > 2:
                    raise OSError

        self._write_error_test(OSError, BadList())

        class BadItem:
            def __str__(self) -> str:
                raise OSError

        self._write_error_test(OSError, [BadItem()])

    def test_write_bigfield(self) -> None:
        bigstring = "X" * 50000
        self._write_test(
            [bigstring, bigstring], "%s,%s" % (bigstring, bigstring)
        )

    def test_write_quoting(self) -> None:
        self._write_test(["a", 1, "p,q"], 'a,1,"p,q"')
        self._write_error_test(
            clevercsv.Error, ["a", 1, "p,q"], quoting=clevercsv.QUOTE_NONE
        )
        self._write_test(
            ["a", 1, "p,q"], 'a,1,"p,q"', quoting=clevercsv.QUOTE_MINIMAL
        )
        self._write_test(
            ["a", 1, "p,q"], '"a",1,"p,q"', quoting=clevercsv.QUOTE_NONNUMERIC
        )
        self._write_test(
            ["a", 1, "p,q"], '"a","1","p,q"', quoting=clevercsv.QUOTE_ALL
        )
        self._write_test(
            ["a\nb", 1], '"a\nb","1"', quoting=clevercsv.QUOTE_ALL
        )

    def test_write_simpledialect(self) -> None:
        self._write_test(
            ["a", 1, "p,q"],
            "a,1,|p,q|",
            dialect=SimpleDialect(delimiter=",", quotechar="|", escapechar=""),
        )

    def test_write_csv_dialect(self) -> None:
        self._write_test(
            ["a", 1, "p,q"],
            'a,1,"p,q"',
            dialect="excel",
        )
        self._write_test(
            ["a", 1, "p,q"],
            '"a","1","p,q"',
            dialect=csv.unix_dialect,
        )
        self._write_test(
            [1, 2, 3],
            "1\t2\t3",
            dialect=clevercsv.excel_tab,
        )


if __name__ == "__main__":
    unittest.main()



import unittest

import clevercsv


class CNamingTestCase(unittest.TestCase):
    def test_name_cabstraction_module(self) -> None:
        self.assertEqual(
            clevercsv.cabstraction.__name__, "clevercsv.cabstraction"
        )

    def test_name_cparser_module(self) -> None:
        self.assertEqual(clevercsv.cparser.__name__, "clevercsv.cparser")

    def test_name_cparser_error(self) -> None:
        self.assertEqual(
            clevercsv.cparser.Error.__module__, "clevercsv.cparser"
        )

    def test_name_cparser_parser(self) -> None:
        self.assertEqual(
            clevercsv.cparser.Parser.__module__, "clevercsv.cparser"
        )



# -*- coding: utf-8 -*-

"""
Common functions for dealing with escape characters.

Author: Gertjan van den Burg
Date: 2018-11-06
"""

import codecs
import sys
import unicodedata

from typing import Iterable
from typing import Optional
from typing import Set

#: Set of default characters to *never* consider as escape character
DEFAULT_BLOCK_CHARS: Set[str] = set(
    [
        "!",
        "?",
        '"',
        "'",
        ".",
        ",",
        ";",
        ":",
        "%",
        "*",
        "&",
        "#",
    ]
)

#: Set of characters in the Unicode "Po" category
UNICODE_PO_CHARS: Set[str] = set(
    [
        c
        for c in map(chr, range(sys.maxunicode + 1))
        if unicodedata.category(c) == "Po"
    ]
)


def is_potential_escapechar(
    char: str, encoding: str, block_char: Optional[Iterable[str]] = None
) -> bool:
    """Check if a character is a potential escape character.

    A character is considered a potential escape character if it is in the
    "Punctuation, Other" Unicode category and not in the list of blocked
    characters.

    Parameters
    ----------
    char: str
        The character to check

    encoding : str
        The encoding of the character

    block_char : Optional[Iterable[str]]
        Characters that are in the Punctuation Other category but that should
        not be considered as escape character. If None, the default set is
        used, which is defined in :py:data:`DEFAULT_BLOCK_CHARS`.

    Returns
    -------
    is_escape : bool
        Whether the character is considered a potential escape or not.

    """
    if encoding.lower() in set(["utf-8", "ascii"]):
        uchar = char
    else:
        uchar = codecs.decode(bytes(char, encoding), encoding=encoding)

    block_chars = (
        DEFAULT_BLOCK_CHARS if block_char is None else set(block_char)
    )
    if uchar in UNICODE_PO_CHARS and uchar not in block_chars:
        return True
    return False



# -*- coding: utf-8 -*-

from __future__ import annotations

from typing import Final
from typing import Generic
from typing import Iterable
from typing import List
from typing import Literal
from typing import Optional
from typing import Tuple
from typing import TypeVar
from typing import overload

_T = TypeVar("_T")

class Parser(Generic[_T]):
    _return_quoted: Final[bool]

    @overload
    def __init__(
        self: Parser[List[Tuple[str, bool]]],
        delimiter: Optional[str] = "",
        quotechar: Optional[str] = "",
        escapechar: Optional[str] = "",
        field_limit: Optional[int] = 128 * 1024,
        strict: Optional[bool] = False,
        return_quoted: Literal[True] = ...,
    ) -> None: ...
    @overload
    def __init__(
        self: Parser[List[str]],
        delimiter: Optional[str] = "",
        quotechar: Optional[str] = "",
        escapechar: Optional[str] = "",
        field_limit: Optional[int] = 128 * 1024,
        strict: Optional[bool] = False,
        return_quoted: Literal[False] = ...,
    ) -> None: ...
    @overload
    def __init__(
        self,
        data: Iterable[str],
        delimiter: Optional[str] = "",
        quotechar: Optional[str] = "",
        escapechar: Optional[str] = "",
        field_limit: Optional[int] = 128 * 1024,
        strict: Optional[bool] = False,
        return_quoted: bool = ...,
    ) -> None: ...
    def __iter__(self) -> "Parser": ...
    def __next__(self) -> _T: ...

class Error(Exception): ...



# -*- coding: utf-8 -*-

"""
Unit tests for the CSV reader.

Author: Gertjan van den Burg

"""

import csv
import json
import unittest

from io import StringIO

from typing import Any
from typing import Iterable
from typing import Iterator
from typing import List

import clevercsv


class ReaderTestCase(unittest.TestCase):
    def _read_test(
        self, input: Iterable[str], expect: List[List[str]], **kwargs: Any
    ) -> None:
        reader = clevercsv.reader(input, **kwargs)
        result = list(reader)
        self.assertEqual(result, expect)

    def test_read_oddinputs(self) -> None:
        self._read_test([], [])
        self._read_test([""], [[]])
        self.assertRaises(
            clevercsv.Error, self._read_test, ['"ab"c'], None, strict=1
        )
        self.assertRaises(
            clevercsv.Error, self._read_test, ["ab\0c"], None, strict=1
        )
        # This commented test doesn't make sense for us, because we assert that
        # a field is only a quoted field if it begins *and* ends with a quote
        # character (except for the last field of a file, which only needs to
        # end in a quotechar if we're not strict).
        # self._read_test(['"ab"c'], [["abc"]], doublequote=0)
        self.assertRaises(clevercsv.Error, self._read_test, [b"ab\0c"], None)

    def test_read_eol(self) -> None:
        self._read_test(["a,b"], [["a", "b"]])
        self._read_test(["a,b\n"], [["a", "b"]])
        self._read_test(["a,b\r\n"], [["a", "b"]])
        self._read_test(["a,b\r"], [["a", "b"]])
        # these tests from CPython don't apply for now.
        self.assertRaises(clevercsv.Error, self._read_test, ["a,b\rc,d"], [])
        self.assertRaises(clevercsv.Error, self._read_test, ["a,b\rc,d"], [])
        self.assertRaises(clevercsv.Error, self._read_test, ["a,b\rc,d"], [])

    def test_read_eof(self) -> None:
        self._read_test(['a,"'], [["a", ""]])
        self._read_test(['"a'], [["a"]])
        # we're not using escape characters in the same way.
        # self._read_test("^", [["\n"]], escapechar="^")

    def test_read_escape(self) -> None:
        # we don't drop the escapechar if it serves no purpose
        # so instead of this:
        # self._read_test("a,\\b,c", [["a", "b", "c"]], escapechar="\\")
        # we do this:
        self._read_test(["a,\\b,c"], [["a", "\\b", "c"]], escapechar="\\")
        self._read_test(["a,b\\,c"], [["a", "b,c"]], escapechar="\\")
        self._read_test(['a,"b,c\\""'], [["a", 'b,c"']], escapechar="\\")
        # the next test also differs from Python
        self._read_test(['a,"b,c"\\'], [["a", "b,c"]], escapechar="\\")

    def test_read_bigfield(self) -> None:
        limit = clevercsv.field_size_limit()
        try:
            size = 500
            bigstring = "X" * size
            bigline = "%s,%s" % (bigstring, bigstring)
            self._read_test([bigline], [[bigstring, bigstring]])
            clevercsv.field_size_limit(size)
            self._read_test([bigline], [[bigstring, bigstring]])
            self.assertEqual(clevercsv.field_size_limit(), size)
            clevercsv.field_size_limit(size=-1)
            self.assertRaises(clevercsv.Error, self._read_test, [bigline], [])
            self.assertRaises(TypeError, clevercsv.field_size_limit, None)
            self.assertRaises(TypeError, clevercsv.field_size_limit, 1, None)
        finally:
            clevercsv.field_size_limit(limit)

    def test_read_linenum(self) -> None:
        r = clevercsv.reader(["line,1", "line,2", "line,3"])
        self.assertEqual(r.line_num, 0)
        self.assertEqual(next(r), ["line", "1"])
        self.assertEqual(r.line_num, 1)
        self.assertEqual(next(r), ["line", "2"])
        self.assertEqual(r.line_num, 2)
        self.assertEqual(next(r), ["line", "3"])
        self.assertEqual(r.line_num, 3)
        self.assertRaises(StopIteration, next, r)
        self.assertEqual(r.line_num, 3)

    def test_with_gen(self) -> None:
        def gen(x: Iterable[str]) -> Iterator[str]:
            for line in x:
                yield line

        r = clevercsv.reader(gen(["line,1", "line,2", "line,3"]))
        self.assertEqual(next(r), ["line", "1"])
        self.assertEqual(next(r), ["line", "2"])
        self.assertEqual(next(r), ["line", "3"])

    def test_simple(self) -> None:
        self._read_test(
            ["A,B,C,D,E"],
            [["A", "B", "C", "D", "E"]],
            delimiter=",",
            quotechar='"',
        )
        self._read_test(
            ["A,B,C,D,E"],
            [["A", "B", "C", "D", "E"]],
            delimiter=",",
            quotechar="",
        )
        self._read_test(["A,B,C,D,E"], [["A,B,C,D,E"]], delimiter="")
        self._read_test(
            ['A,"B",C,D,E'],
            [["A", "B", "C", "D", "E"]],
            delimiter=",",
            quotechar='"',
        )
        self._read_test(
            ['A,"B,C",D,E'],
            [["A", "B,C", "D", "E"]],
            delimiter=",",
            quotechar='"',
        )
        self._read_test(
            ['A,"B,C",D,E'],
            [["A", '"B', 'C"', "D", "E"]],
            delimiter=",",
            quotechar="",
        )
        self._read_test(
            ['"A","B","C",,,,'],
            [['"A"', '"B"', '"C"', "", "", "", ""]],
            delimiter=",",
            quotechar="",
        )

    def test_no_delim(self) -> None:
        self._read_test(
            ['A"B"C', 'A"B""C""D"'],
            [['A"B"C'], ['A"B""C""D"']],
            delimiter="",
            quotechar="",
        )
        self._read_test(
            ['A"B\r\nB"C', 'D"E"F', "G"],
            [['A"B\r\nB"C'], ['D"E"F'], ["G"]],
            delimiter="",
            quotechar='"',
        )
        self._read_test(
            ['A"B\nB"C', 'D"E"F', "G"],
            [['A"B\nB"C'], ['D"E"F'], ["G"]],
            delimiter="",
            quotechar='"',
        )
        self._read_test(
            ['A"B\nB\rB"C', 'D"E"F', "G"],
            [['A"B\nB\rB"C'], ['D"E"F'], ["G"]],
            delimiter="",
            quotechar='"',
        )

    def test_github_issue_130(self) -> None:
        # fmt: off
        data = """sku,features,attributes
22221,"[{""key"":""heel_height"",""value"":""Ulttra High (4\\""+)""}]","11,room"
"""
        # fmt: on

        # NOTE we set the escapechar to None to correctly carry over the `\`
        # escape character
        rows_csv = list(
            csv.reader(
                StringIO(data), delimiter=",", quotechar='"', escapechar=None
            )
        )
        expected = [
            [
                "sku",
                "features",
                "attributes",
            ],
            [
                "22221",
                '[{"key":"heel_height","value":"Ulttra High (4\\"+)"}]',
                "11,room",
            ],
        ]
        self.assertSequenceEqual(expected, rows_csv)

        rows_clevercsv = list(
            clevercsv.reader(
                StringIO(data), delimiter=",", quotechar='"', escapechar=""
            )
        )
        self.assertSequenceEqual(expected, rows_clevercsv)

        # Ensure we can parse the JSON as intended.
        cell = json.loads(expected[1][1])
        self.assertEqual(
            cell, [{"key": "heel_height", "value": 'Ulttra High (4"+)'}]
        )


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-

"""
Code for computing the type score.

Author: Gertjan van den Burg

"""

import json

from typing import Dict
from typing import List
from typing import Optional
from typing import Pattern

from ._regexes import DEFAULT_TYPE_REGEXES
from .cparser_util import parse_string
from .dialect import SimpleDialect

DEFAULT_EPS_TYPE: float = 1e-10


class TypeDetector:
    def __init__(
        self,
        patterns: Optional[Dict[str, Pattern[str]]] = None,
        strip_whitespace: bool = True,
    ) -> None:
        self.patterns = patterns or DEFAULT_TYPE_REGEXES.copy()
        self.strip_whitespace = strip_whitespace
        self._register_type_tests()

    def _register_type_tests(self) -> None:
        self._type_tests = [
            ("empty", self.is_empty),
            ("url", self.is_url),
            ("email", self.is_email),
            ("ipv4", self.is_ipv4),
            ("number", self.is_number),
            ("time", self.is_time),
            ("percentage", self.is_percentage),
            ("currency", self.is_currency),
            ("unix_path", self.is_unix_path),
            ("nan", self.is_nan),
            ("date", self.is_date),
            ("datetime", self.is_datetime),
            ("unicode_alphanum", self.is_unicode_alphanum),
            ("bytearray", self.is_bytearray),
            ("json", self.is_json_obj),
        ]

    def list_known_types(self) -> List[str]:
        return [tt[0] for tt in self._type_tests]

    def is_known_type(self, cell: str, is_quoted: bool = False) -> bool:
        return self.detect_type(cell, is_quoted=is_quoted) is not None

    def detect_type(self, cell: str, is_quoted: bool = False) -> Optional[str]:
        cell = cell.strip() if self.strip_whitespace else cell
        for name, func in self._type_tests:
            if func(cell, is_quoted=is_quoted):
                return name
        return None

    def _run_regex(self, cell: str, patname: str) -> bool:
        cell = cell.strip() if self.strip_whitespace else cell
        pat = self.patterns.get(patname, None)
        assert pat is not None
        match = pat.fullmatch(cell)
        return match is not None

    def is_number(self, cell: str, is_quoted: bool = False) -> bool:
        if cell == "":
            return False
        if self._run_regex(cell, "number_1"):
            return True
        if self._run_regex(cell, "number_2"):
            return True
        if self._run_regex(cell, "number_3"):
            return True
        return False

    def is_ipv4(self, cell: str, is_quoted: bool = False) -> bool:
        return self._run_regex(cell, "ipv4")

    def is_url(self, cell: str, is_quoted: bool = False) -> bool:
        return self._run_regex(cell, "url")

    def is_email(self, cell: str, is_quoted: bool = False) -> bool:
        return self._run_regex(cell, "email")

    def is_unicode_alphanum(self, cell: str, is_quoted: bool = False) -> bool:
        if is_quoted:
            return self._run_regex(cell, "unicode_alphanum_quoted")
        return self._run_regex(cell, "unicode_alphanum")

    def is_date(self, cell: str, is_quoted: bool = False) -> bool:
        # This function assumes the cell is not a number.
        cell = cell.strip() if self.strip_whitespace else cell
        if not cell:
            return False
        if not cell[0].isdigit():
            return False
        return self._run_regex(cell, "date")

    def is_time(self, cell: str, is_quoted: bool = False) -> bool:
        cell = cell.strip() if self.strip_whitespace else cell
        if not cell:
            return False
        if not cell[0].isdigit():
            return False
        return (
            self._run_regex(cell, "time_hmm")
            or self._run_regex(cell, "time_hhmm")
            or self._run_regex(cell, "time_hhmmss")
            or self._run_regex(cell, "time_hhmmsszz")
        )

    def is_empty(self, cell: str, is_quoted: bool = False) -> bool:
        return cell == ""

    def is_percentage(self, cell: str, is_quoted: bool = False) -> bool:
        return cell.endswith("%") and self.is_number(cell.rstrip("%"))

    def is_currency(self, cell: str, is_quoted: bool = False) -> bool:
        pat = self.patterns.get("currency", None)
        assert pat is not None
        m = pat.fullmatch(cell)
        if m is None:
            return False
        grp = m.group(1)
        if not self.is_number(grp):
            return False
        return True

    def is_datetime(self, cell: str, is_quoted: bool = False) -> bool:
        # Takes care of cells with '[date] [time]' and '[date]T[time]' (iso)
        if not cell:
            return False

        if not cell[0].isdigit():
            return False

        if " " in cell:
            parts = cell.split(" ")
            if len(parts) > 2:
                return False
            return self.is_date(parts[0]) and self.is_time(parts[1])
        elif "T" in cell:
            parts = cell.split("T")
            if len(parts) > 2:
                return False
            isdate = self.is_date(parts[0])
            if not isdate:
                return False
            # [date]T[time] or [date]T[time]Z
            if parts[1].endswith("Z") and self.is_time(parts[1][:-1]):
                return True
            if self.is_time(parts[1]):
                return True
            # [date]T[time][+-][time]
            if "+" in parts[1]:
                subparts = parts[1].split("+")
                istime1 = self.is_time(subparts[0])
                if not istime1:
                    return False
                istime2 = self.is_time(subparts[1])
                if istime2:
                    return True
                if self._run_regex(subparts[1], "time_HHMM"):
                    return True
                if self._run_regex(subparts[1], "time_HH"):
                    return True
            elif "-" in parts[1]:
                subparts = parts[1].split("-")
                istime1 = self.is_time(subparts[0])
                if not istime1:
                    return False
                istime2 = self.is_time(subparts[1])
                if istime2:
                    return True
                if self._run_regex(subparts[1], "time_HHMM"):
                    return True
                if self._run_regex(subparts[1], "time_HH"):
                    return True
        return False

    def is_nan(self, cell: str, is_quoted: bool = False) -> bool:
        if cell.lower() in ["n/a", "na", "nan"]:
            return True
        return False

    def is_unix_path(self, cell: str, is_quoted: bool = False) -> bool:
        return self._run_regex(cell, "unix_path")

    def is_bytearray(self, cell: str, is_quoted: bool = False) -> bool:
        return cell.startswith("bytearray(b") and cell.endswith(")")

    def is_json_obj(self, cell: str, is_quoted: bool = False) -> bool:
        if not (cell.startswith("{") and cell.endswith("}")):
            return False
        try:
            _ = json.loads(cell)
        except json.JSONDecodeError:
            return False
        return True


def gen_known_type(cells):
    """
    Utility that yields a generator over whether or not the provided cells are
    of a known type or not.
    """
    td = TypeDetector()
    for cell in cells:
        yield td.is_known_type(cell)


def type_score(
    data: str, dialect: SimpleDialect, eps: float = DEFAULT_EPS_TYPE
) -> float:
    """
    Compute the type score as the ratio of cells with a known type.

    Parameters
    ----------
    data: str
        the data as a single string

    dialect: SimpleDialect
        the dialect to use

    eps: float
        the minimum value of the type score

    Returns
    -------
    type_score: float
        The computed type score

    """
    total = 0
    known = 0
    td = TypeDetector()
    for row in parse_string(data, dialect, return_quoted=True):
        for cell, is_quoted in row:
            total += 1
            known += td.is_known_type(cell, is_quoted=is_quoted)
    if total == 0:
        return eps
    return max(eps, known / total)



from typing import Any

from pandas._config import describe_option as describe_option
from pandas._config import get_option as get_option
from pandas._config import option_context as option_context
from pandas._config import options as options
from pandas._config import reset_option as reset_option
from pandas._config import set_option as set_option
from pandas.core.api import NA as NA
from pandas.core.api import BooleanDtype as BooleanDtype
from pandas.core.api import Categorical as Categorical
from pandas.core.api import CategoricalDtype as CategoricalDtype
from pandas.core.api import CategoricalIndex as CategoricalIndex
from pandas.core.api import DataFrame as DataFrame
from pandas.core.api import DateOffset as DateOffset
from pandas.core.api import DatetimeIndex as DatetimeIndex
from pandas.core.api import DatetimeTZDtype as DatetimeTZDtype
from pandas.core.api import Flags as Flags
from pandas.core.api import Float32Dtype as Float32Dtype
from pandas.core.api import Float64Dtype as Float64Dtype
from pandas.core.api import Float64Index as Float64Index
from pandas.core.api import Grouper as Grouper
from pandas.core.api import Index as Index
from pandas.core.api import IndexSlice as IndexSlice
from pandas.core.api import Int8Dtype as Int8Dtype
from pandas.core.api import Int16Dtype as Int16Dtype
from pandas.core.api import Int32Dtype as Int32Dtype
from pandas.core.api import Int64Dtype as Int64Dtype
from pandas.core.api import Int64Index as Int64Index
from pandas.core.api import Interval as Interval
from pandas.core.api import IntervalDtype as IntervalDtype
from pandas.core.api import IntervalIndex as IntervalIndex
from pandas.core.api import MultiIndex as MultiIndex
from pandas.core.api import NamedAgg as NamedAgg
from pandas.core.api import NaT as NaT
from pandas.core.api import Period as Period
from pandas.core.api import PeriodDtype as PeriodDtype
from pandas.core.api import PeriodIndex as PeriodIndex
from pandas.core.api import RangeIndex as RangeIndex
from pandas.core.api import Series as Series
from pandas.core.api import StringDtype as StringDtype
from pandas.core.api import Timedelta as Timedelta
from pandas.core.api import TimedeltaIndex as TimedeltaIndex
from pandas.core.api import Timestamp as Timestamp
from pandas.core.api import UInt8Dtype as UInt8Dtype
from pandas.core.api import UInt16Dtype as UInt16Dtype
from pandas.core.api import UInt32Dtype as UInt32Dtype
from pandas.core.api import UInt64Dtype as UInt64Dtype
from pandas.core.api import UInt64Index as UInt64Index
from pandas.core.api import array as array
from pandas.core.api import bdate_range as bdate_range
from pandas.core.api import date_range as date_range
from pandas.core.api import factorize as factorize
from pandas.core.api import interval_range as interval_range
from pandas.core.api import isna as isna
from pandas.core.api import isnull as isnull
from pandas.core.api import notna as notna
from pandas.core.api import notnull as notnull
from pandas.core.api import period_range as period_range
from pandas.core.api import set_eng_float_format as set_eng_float_format
from pandas.core.api import timedelta_range as timedelta_range
from pandas.core.api import to_datetime as to_datetime
from pandas.core.api import to_numeric as to_numeric
from pandas.core.api import to_timedelta as to_timedelta
from pandas.core.api import unique as unique
from pandas.core.api import value_counts as value_counts
from pandas.core.arrays.sparse import SparseDtype as SparseDtype
from pandas.core.computation.api import eval as eval
from pandas.core.reshape.api import concat as concat
from pandas.core.reshape.api import crosstab as crosstab
from pandas.core.reshape.api import cut as cut
from pandas.core.reshape.api import get_dummies as get_dummies
from pandas.core.reshape.api import lreshape as lreshape
from pandas.core.reshape.api import melt as melt
from pandas.core.reshape.api import merge as merge
from pandas.core.reshape.api import merge_asof as merge_asof
from pandas.core.reshape.api import merge_ordered as merge_ordered
from pandas.core.reshape.api import pivot as pivot
from pandas.core.reshape.api import pivot_table as pivot_table
from pandas.core.reshape.api import qcut as qcut
from pandas.core.reshape.api import wide_to_long as wide_to_long
from pandas.io.api import ExcelFile as ExcelFile
from pandas.io.api import ExcelWriter as ExcelWriter
from pandas.io.api import HDFStore as HDFStore
from pandas.io.api import read_clipboard as read_clipboard
from pandas.io.api import read_csv as read_csv
from pandas.io.api import read_excel as read_excel
from pandas.io.api import read_feather as read_feather
from pandas.io.api import read_fwf as read_fwf
from pandas.io.api import read_gbq as read_gbq
from pandas.io.api import read_hdf as read_hdf
from pandas.io.api import read_html as read_html
from pandas.io.api import read_json as read_json
from pandas.io.api import read_orc as read_orc
from pandas.io.api import read_parquet as read_parquet
from pandas.io.api import read_pickle as read_pickle
from pandas.io.api import read_sas as read_sas
from pandas.io.api import read_spss as read_spss
from pandas.io.api import read_sql as read_sql
from pandas.io.api import read_sql_query as read_sql_query
from pandas.io.api import read_sql_table as read_sql_table
from pandas.io.api import read_stata as read_stata
from pandas.io.api import read_table as read_table
from pandas.io.api import to_pickle as to_pickle
from pandas.tseries import offsets as offsets
from pandas.tseries.api import infer_freq as infer_freq
from pandas.util._print_versions import show_versions as show_versions
from pandas.util._tester import test as test

__docformat__: str
hard_dependencies: Any
missing_dependencies: Any
module: Any
v: Any
__git_version__: Any

def __getattr__(name: Any): ...

# __doc__: str



# -*- coding: utf-8 -*-

"""Code for dealing with optional dependencies

The functionality in this file is largely based on similar functionality in the 
Pandas library.

Author: G.J.J van den Burg
Copyright: 2020, The Alan Turing Institute
License: See LICENSE file.

"""

import importlib

from types import ModuleType

from typing import Dict
from typing import List
from typing import NamedTuple
from typing import Optional

from packaging.version import Version


class OptionalDependency(NamedTuple):
    import_name: str
    package_name: str
    min_version: str


# update this when changing setup.py
OPTIONAL_DEPENDENCIES: List[OptionalDependency] = [
    OptionalDependency("tabview", "tabview", "1.4"),
    OptionalDependency("pandas", "pandas", "0.24.1"),
    OptionalDependency("cchardet", "faust-cchardet", "2.1.18"),
    OptionalDependency("wilderness", "wilderness", "0.1.5"),
]


def import_optional_dependency(
    name: str, raise_on_missing: bool = True
) -> Optional[ModuleType]:
    """
    Import an optional dependency.

    This function is modelled on a similar function in the Pandas library.

    Parameters
    ----------
    name : str
        Name of the module to import

    raise_on_missing : bool
        Whether to raise an error when the package is missing or to simply
        return None.

    Returns
    -------
    module : module
        The module if importing was successful, None if
        :attr:`raise_on_missing` is False.

    Raises
    ------
    ImportError
        When a module can't be imported and :attr:`raise_on_missing` is True.

    """
    msg = (
        f"\nOptional dependency '{name}' is missing. You can install it using "
        "pip or conda, or you can install CleverCSV with all of its optional "
        "dependencies by running: pip install clevercsv[full]"
    )
    try:
        module = importlib.import_module(name)
    except ImportError:
        if raise_on_missing:
            raise ImportError(msg) from None
        else:
            return None

    opt_dependencies: Dict[str, OptionalDependency] = {
        d.import_name: d for d in OPTIONAL_DEPENDENCIES
    }

    dependency = opt_dependencies.get(name)
    if dependency is None:
        raise ImportError(f"No known optional dependency with name: {name}")

    version = getattr(module, "__version__", None)
    if version is None:
        return module

    if Version(version) < Version(dependency.min_version):
        msg = (
            f"CleverCSV requires version '{dependency.min_version}' or newer "
            f"for optional dependency '{dependency.package_name}'. Please "
            "update the package or install CleverCSV with all its optional "
            "dependencies using: pip install clevercsv[full]"
        )
        raise ImportError(msg)

    return module



# -*- coding: utf-8 -*-

from typing import Optional

def base_abstraction(
    data: str,
    delimiter: Optional[str],
    quotechar: Optional[str],
    escapechar: Optional[str],
) -> str: ...
def c_merge_with_quotechar(data: str) -> str: ...



# -*- coding: utf-8 -*-

"""
Wrappers for some loading/saving functionality.

Author: Gertjan van den Burg

"""
from __future__ import annotations

import os
import warnings

from typing import TYPE_CHECKING
from typing import Any
from typing import Iterable
from typing import Iterator
from typing import List
from typing import Mapping
from typing import Optional
from typing import TypeVar

from ._optional import import_optional_dependency
from .detect import Detector
from .dialect import SimpleDialect
from .dict_read_write import DictReader
from .dict_read_write import DictWriter
from .encoding import get_encoding
from .exceptions import NoDetectionResult
from .read import reader
from .write import writer

if TYPE_CHECKING:
    import pandas as pd

    from ._types import FileDescriptorOrPath
    from ._types import _DialectLike
    from ._types import _DictReadMapping

_T = TypeVar("_T")


def stream_dicts(
    filename: FileDescriptorOrPath,
    dialect: Optional[_DialectLike] = None,
    encoding: Optional[str] = None,
    num_chars: Optional[int] = None,
    verbose: bool = False,
) -> Iterator["_DictReadMapping"]:
    """Read a CSV file as a generator over dictionaries

    This function streams the rows of the CSV file as dictionaries. The keys of
    the dictionaries are assumed to be in the first row of the CSV file. The
    dialect will be detected automatically, unless it is provided.

    Parameters
    ----------
    filename : str
        Path of the CSV file

    dialect : str, SimpleDialect, or csv.Dialect object
        If the dialect is known, it can be provided here. This function uses
        the Clevercsv :class:`clevercsv.DictReader` object, which supports
        various dialect types (string, SimpleDialect, or csv.Dialect). If None,
        the dialect will be detected.

    encoding : str
        The encoding of the file. If None, it is detected.

    num_chars : int
        Number of characters to use to detect the dialect. If None, use the
        entire file.

        Note that using less than the entire file will speed up detection, but
        can reduce the accuracy of the detected dialect.

    verbose: bool
        Whether or not to show detection progress.

    Returns
    -------
    rows: generator
        Returns file as a generator over rows as dictionaries.

    Raises
    ------
    NoDetectionResult
        When the dialect detection fails.

    """
    if encoding is None:
        encoding = get_encoding(filename)
    with open(filename, "r", newline="", encoding=encoding) as fid:
        if dialect is None:
            data = fid.read(num_chars) if num_chars else fid.read()
            dialect = Detector().detect(data, verbose=verbose)
            fid.seek(0)

        if dialect is None:
            raise NoDetectionResult

        reader: DictReader = DictReader(fid, dialect=dialect)
        for row in reader:
            yield row


def read_dicts(
    filename: "FileDescriptorOrPath",
    dialect: Optional["_DialectLike"] = None,
    encoding: Optional[str] = None,
    num_chars: Optional[int] = None,
    verbose: bool = False,
) -> List["_DictReadMapping"]:
    """Read a CSV file as a list of dictionaries

    This function returns the rows of the CSV file as a list of dictionaries.
    The keys of the dictionaries are assumed to be in the first row of the CSV
    file. The dialect will be detected automatically, unless it is provided.

    Parameters
    ----------
    filename : str
        Path of the CSV file

    dialect : str, SimpleDialect, or csv.Dialect object
        If the dialect is known, it can be provided here. This function uses
        the Clevercsv :class:`clevercsv.DictReader` object, which supports
        various dialect types (string, SimpleDialect, or csv.Dialect). If None,
        the dialect will be detected.

    encoding : str
        The encoding of the file. If None, it is detected.

    num_chars : int
        Number of characters to use to detect the dialect. If None, use the
        entire file.

        Note that using less than the entire file will speed up detection, but
        can reduce the accuracy of the detected dialect.

    verbose: bool
        Whether or not to show detection progress.

    Returns
    -------
    rows: list
        Returns rows of the file as a list of dictionaries.

    Raises
    ------
    NoDetectionResult
        When the dialect detection fails.

    """
    return list(
        stream_dicts(
            filename,
            dialect=dialect,
            encoding=encoding,
            num_chars=num_chars,
            verbose=verbose,
        )
    )


def read_table(
    filename: "FileDescriptorOrPath",
    dialect: Optional["_DialectLike"] = None,
    encoding: Optional[str] = None,
    num_chars: Optional[int] = None,
    verbose: bool = False,
) -> List[List[str]]:
    """Read a CSV file as a table (a list of lists)

    This is a convenience function that reads a CSV file and returns the data
    as a list of lists (= rows). The dialect will be detected automatically,
    unless it is provided.

    Parameters
    ----------
    filename: str
        Path of the CSV file

    dialect: str, SimpleDialect, or csv.Dialect object
        If the dialect is known, it can be provided here. This function uses
        the CleverCSV :class:`clevercsv.reader` object, which supports various
        dialect types (string, SimpleDialect, or csv.Dialect). If None, the
        dialect will be detected.

    encoding : str
        The encoding of the file. If None, it is detected.

    num_chars : int
        Number of characters to use to detect the dialect. If None, use the
        entire file.

        Note that using less than the entire file will speed up detection, but
        can reduce the accuracy of the detected dialect.

    verbose: bool
        Whether or not to show detection progress.

    Returns
    -------
    rows: list
        Returns rows as a list of lists.

    Raises
    ------
    NoDetectionResult
        When the dialect detection fails.

    """
    return list(
        stream_table(
            filename,
            dialect=dialect,
            encoding=encoding,
            num_chars=num_chars,
            verbose=verbose,
        )
    )


def stream_table(
    filename: "FileDescriptorOrPath",
    dialect: Optional["_DialectLike"] = None,
    encoding: Optional[str] = None,
    num_chars: Optional[int] = None,
    verbose: bool = False,
) -> Iterator[List[str]]:
    """Read a CSV file as a generator over rows of a table

    This is a convenience function that reads a CSV file and returns the data
    as a generator of rows. The dialect will be detected automatically, unless
    it is provided.

    Parameters
    ----------
    filename: str
        Path of the CSV file

    dialect: str, SimpleDialect, or csv.Dialect object
        If the dialect is known, it can be provided here. This function uses
        the CleverCSV :class:`clevercsv.reader` object, which supports various
        dialect types (string, SimpleDialect, or csv.Dialect). If None, the
        dialect will be detected.

    encoding : str
        The encoding of the file. If None, it is detected.

    num_chars : int
        Number of characters to use to detect the dialect. If None, use the
        entire file.

        Note that using less than the entire file will speed up detection, but
        can reduce the accuracy of the detected dialect.

    verbose: bool
        Whether or not to show detection progress.

    Returns
    -------
    rows: generator
        Returns file as a generator over rows.

    Raises
    ------
    NoDetectionResult
        When the dialect detection fails.

    """
    if encoding is None:
        encoding = get_encoding(filename)
    with open(filename, "r", newline="", encoding=encoding) as fid:
        if dialect is None:
            data = fid.read(num_chars) if num_chars else fid.read()
            dialect = Detector().detect(data, verbose=verbose)
            if dialect is None:
                raise NoDetectionResult()
            fid.seek(0)
        r = reader(fid, dialect)
        yield from r


def read_dataframe(
    filename: "FileDescriptorOrPath",
    *args: Any,
    num_chars: Optional[int] = None,
    **kwargs: Any,
) -> pd.DataFrame:
    """Read a CSV file to a Pandas dataframe

    This function uses CleverCSV to detect the dialect, and then passes this to
    the ``read_csv`` function in pandas. Additional arguments and keyword
    arguments are passed to ``read_csv`` as well.

    Parameters
    ----------

    filename: str
        The filename of the CSV file. At the moment, only local files are
        supported.

    *args:
        Additional arguments for the ``pandas.read_csv`` function.

    num_chars: int
        Number of characters to use for dialect detection. If None, use the
        entire file.

        Note that using less than the entire file will speed up detection, but
        can reduce the accuracy of the detected dialect.

    **kwargs:
        Additional keyword arguments for the ``pandas.read_csv`` function. You
        can specify the file encoding here if needed, and it will be used
        during dialect detection.

    """
    if not (os.path.exists(filename) and os.path.isfile(filename)):
        raise ValueError("Filename must be a regular file")
    pd = import_optional_dependency("pandas")
    assert pd is not None

    # Use provided encoding or detect it, and record it for pandas
    enc = kwargs.get("encoding") or get_encoding(filename)
    kwargs["encoding"] = enc

    with open(filename, "r", newline="", encoding=enc) as fid:
        data = fid.read(num_chars) if num_chars else fid.read()
        dialect = Detector().detect(data)

    if dialect is None:
        raise NoDetectionResult

    csv_dialect = dialect.to_csv_dialect()

    # This is used to catch pandas' warnings when a dialect is supplied.
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message="^Conflicting values for .*",
            category=pd.errors.ParserWarning,
        )
        df = pd.read_csv(filename, *args, dialect=csv_dialect, **kwargs)
    return df


def detect_dialect(
    filename: "FileDescriptorOrPath",
    num_chars: Optional[int] = None,
    encoding: Optional[str] = None,
    verbose: bool = False,
    method: str = "auto",
    skip: bool = True,
) -> Optional[SimpleDialect]:
    """Detect the dialect of a CSV file

    This is a utility function that simply returns the detected dialect of a
    given CSV file.

    Parameters
    ----------
    filename : str
        The filename of the CSV file.

    num_chars : int
        Number of characters to read for the detection. If None, the entire
        file will be read. Note that limiting the number of characters can
        reduce the accuracy of the detected dialect.

    encoding : str
        The file encoding of the CSV file. If None, it is detected.

    verbose : bool
        Enable verbose mode during detection.

    method : str
        Dialect detection method to use. Either 'normal' for normal form
        detection, 'consistency' for the consistency measure, or 'auto' for
        first normal and then consistency.

    skip : bool
        Skip computation of the type score for dialects with a low pattern
        score.

    Returns
    -------
    dialect : Optional[SimpleDialect]
        The detected dialect as a :class:`SimpleDialect`, or None if detection
        failed.

    """
    enc = encoding or get_encoding(filename)
    with open(filename, "r", newline="", encoding=enc) as fp:
        data = fp.read(num_chars) if num_chars else fp.read()
        dialect = Detector().detect(
            data, verbose=verbose, method=method, skip=skip
        )
    return dialect


def write_table(
    table: Iterable[Iterable[Any]],
    filename: "FileDescriptorOrPath",
    dialect: "_DialectLike" = "excel",
    transpose: bool = False,
    encoding: Optional[str] = None,
) -> None:
    """Write a table (a list of lists) to a file

    This is a convenience function for writing a table to a CSV file. If the
    table has no rows, no output file is created.

    Parameters
    ----------
    table : list
        A table as a list of lists. The table must have the same number of
        cells in each row (taking the :attr:`transpose` flag into account).

    filename : str
        The filename of the CSV file to write the table to.

    dialect : SimpleDialect or csv.Dialect
        The dialect to use. The default is the 'excel' dialect, which
        corresponds to RFC4180. This is done to encourage more standardized CSV
        files.

    transpose : bool
        Transpose the table before writing.

    encoding : str
        Encoding to use to write the data to the file. Note that the default
        encoding is platform dependent, which ensures compatibility with the
        Python open() function. It thus defaults to
        `locale.getpreferredencoding()`.

    Raises
    ------
    ValueError:
        When the length of the rows is not constant.

    """
    if not table:
        return

    if transpose:
        list_table = list(map(list, zip(*table)))
    else:
        list_table = list(map(list, table))

    if len(set(map(len, list_table))) > 1:
        raise ValueError("Table doesn't have constant row length.")

    with open(filename, "w", newline="", encoding=encoding) as fp:
        w = writer(fp, dialect=dialect)
        w.writerows(list_table)


def write_dicts(
    items: Iterable[Mapping[_T, Any]],
    filename: "FileDescriptorOrPath",
    dialect: "_DialectLike" = "excel",
    encoding: Optional[str] = None,
) -> None:
    """Write a list of dicts to a file

    This is a convenience function to write dicts to a file. The header is
    extracted from the keys of the first item, so an OrderedDict is recommended
    to control the order of the headers in the output. If the list of items is
    empty, no output file is created.

    Parameters
    ----------
    items : list
        List of dicts to export

    filename : str
        The filename of the CSV file to write the table to

    dialect : str, SimpleDialect, or csv.Dialect
        The dialect to use. The default is the 'excel' dialect, which
        corresponds to RFC4180.

    encoding : str
        Encoding to use to write the data to the file. Note that the default
        encoding is platform dependent, which ensures compatibility with the
        Python open() function. It thus defaults to
        `locale.getpreferredencoding()`.

    """
    if not items:
        return

    iterator = iter(items)
    try:
        first = next(iterator)
    except StopIteration:
        return

    fieldnames = list(first.keys())
    with open(filename, "w", newline="", encoding=encoding) as fp:
        w = DictWriter(fp, fieldnames=fieldnames, dialect=dialect)
        w.writeheader()
        w.writerow(first)
        w.writerows(iterator)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import glob
import io
import os

from setuptools import Command
from setuptools import Extension
from setuptools import find_packages
from setuptools import setup

# Package meta-data.
AUTHOR = "Gertjan van den Burg"
DESCRIPTION = "A Python package for handling messy CSV files"
EMAIL = "gertjanvandenburg@gmail.com"
LICENSE = "MIT"
LICENSE_TROVE = "License :: OSI Approved :: MIT License"
NAME = "clevercsv"
REQUIRES_PYTHON = ">=3.8.0"
URL = "https://github.com/alan-turing-institute/CleverCSV"
VERSION = None

# What packages are required for this module to be executed?
REQUIRED = [
    "chardet>=3.0",
    "regex>=2018.11",
    "packaging>=23.0",
]

# Dependencies only needed for pre-commit
precommit_require = [
    "wilderness>=0.1.5",
]

# When these are changed, update clevercsv/_optional.py accordingly
full_require = [
    *precommit_require,
    "faust-cchardet>=2.1.18",
    "pandas>=1.0.0",
    "tabview>=1.4",
]

docs_require = ["sphinx", "m2r2", "furo"]
test_require = full_require
dev_require = [
    "green",
    # "pythonfuzz",
    "pytest>=2.6",
    "termcolor",
    "mypy",
]

# What packages are optional?
EXTRAS = {
    "full": full_require,
    "docs": docs_require,
    "tests": test_require,
    "dev": docs_require + test_require + dev_require,
    "precommit": precommit_require,
}


class build_manpages(Command):
    description = "Generate manpages"
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        from wilderness import build_manpages

        from clevercsv.console import build_application

        build_manpages(build_application())


# The rest you shouldn't have to touch too much :)
# ------------------------------------------------
# Except, perhaps the License and Trove Classifiers!
# If you do change the License, remember to change the Trove Classifier for that!

here = os.path.abspath(os.path.dirname(__file__))

# Import the README and use it as the long-description.
# Note: this will only work if 'README.md' is present in your MANIFEST.in file!
try:
    with io.open(os.path.join(here, "README.md"), encoding="utf-8") as f:
        long_description = "\n" + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION

# Load the package's __version__.py module as a dictionary.
about = {}
if not VERSION:
    project_slug = NAME.lower().replace("-", "_").replace(" ", "_")
    with open(os.path.join(here, project_slug, "__version__.py")) as f:
        exec(f.read(), about)
else:
    about["__version__"] = VERSION

# Where the magic happens:
setup(
    name=NAME,
    version=about["__version__"],
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type="text/markdown",
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    packages=find_packages(
        exclude=["tests", "*.tests", "*.tests.*", "tests.*"]
    ),
    install_requires=REQUIRED,
    extras_require=EXTRAS,
    include_package_data=True,
    package_data={"clevercsv": ["py.typed"]},
    license=LICENSE,
    ext_modules=[
        Extension("clevercsv.cparser", sources=["src/cparser.c"]),
        Extension("clevercsv.cabstraction", sources=["src/abstraction.c"]),
    ],
    entry_points={"console_scripts": ["clevercsv = clevercsv.__main__:main"]},
    data_files=[("man/man1", glob.glob("man/*.1"))],
    cmdclass={"build_manpages": build_manpages},
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        LICENSE_TROVE,
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
    ],
)



# -*- coding: utf-8 -*-

"""
Unit tests for the wrappers

Author: Gertjan van den Burg

"""

import os
import tempfile
import types
import unittest

from typing import Any
from typing import Dict
from typing import Iterable
from typing import List
from typing import Union

import pandas as pd

from clevercsv import wrappers
from clevercsv import writer
from clevercsv.dialect import SimpleDialect
from clevercsv.exceptions import NoDetectionResult


class WrappersTestCase(unittest.TestCase):
    def _df_test(
        self, table: List[List[Any]], dialect: SimpleDialect, **kwargs: Any
    ) -> None:
        tmpfd, tmpfname = tempfile.mkstemp(prefix="ccsv_", suffix=".csv")
        tmpid = os.fdopen(tmpfd, "w", encoding=kwargs.get("encoding"))
        w = writer(tmpid, dialect=dialect)
        w.writerows(table)
        tmpid.close()

        exp_df = pd.DataFrame.from_records(table[1:], columns=table[0])
        df = wrappers.read_dataframe(tmpfname)

        try:
            self.assertTrue(df.equals(exp_df))
        finally:
            os.unlink(tmpfname)

    def _write_tmpfile(
        self, table: Iterable[Iterable[Any]], dialect: SimpleDialect
    ) -> str:
        """Write a table to a temporary file using specified dialect"""
        tmpfd, tmpfname = tempfile.mkstemp(prefix="ccsv_", suffix=".csv")
        tmpid = os.fdopen(tmpfd, "w")
        w = writer(tmpid, dialect=dialect)
        w.writerows(table)
        tmpid.close()
        return tmpfname

    def _read_test(
        self, table: Iterable[Iterable[Any]], dialect: SimpleDialect
    ) -> None:
        tmpfname = self._write_tmpfile(table, dialect)
        exp = [list(map(str, r)) for r in table]
        try:
            self.assertEqual(exp, wrappers.read_table(tmpfname))
        finally:
            os.unlink(tmpfname)

    def _stream_test(
        self, table: Iterable[Iterable[str]], dialect: SimpleDialect
    ) -> None:
        tmpfname = self._write_tmpfile(table, dialect)
        exp = [list(map(str, r)) for r in table]
        try:
            out = wrappers.stream_table(tmpfname)
            self.assertTrue(isinstance(out, types.GeneratorType))
            self.assertEqual(exp, list(out))
        finally:
            os.unlink(tmpfname)

    def _read_test_rows(
        self, rows: List[str], expected: List[List[str]]
    ) -> None:
        contents = "\n".join(rows)
        tmpfd, tmpfname = tempfile.mkstemp(prefix="ccsv_", suffix=".csv")
        tmpid = os.fdopen(tmpfd, "w")
        tmpid.write(contents)
        tmpid.close()

        try:
            self.assertEqual(expected, wrappers.read_table(tmpfname))
        finally:
            os.unlink(tmpfname)

    def _stream_test_rows(
        self, rows: Iterable[str], expected: List[List[str]]
    ) -> None:
        contents = "\n".join(rows)
        tmpfd, tmpfname = tempfile.mkstemp(prefix="ccsv_", suffix=".csv")
        tmpid = os.fdopen(tmpfd, "w")
        tmpid.write(contents)
        tmpid.close()

        try:
            out = wrappers.stream_table(tmpfname)
            self.assertTrue(isinstance(out, types.GeneratorType))
            self.assertEqual(expected, list(out))
        finally:
            os.unlink(tmpfname)

    def test_read_dataframe(self) -> None:
        table: List[List[Any]]

        table = [["A", "B", "C"], [1, 2, 3], [4, 5, 6]]
        dialect = SimpleDialect(delimiter=";", quotechar="", escapechar="")
        with self.subTest(name="simple"):
            self._df_test(table, dialect)

        table = [["A,0", "B", "C"], [1, 2, 3], [4, 5, 6]]
        dialect = SimpleDialect(delimiter=",", quotechar="", escapechar="\\")
        with self.subTest(name="escaped"):
            self._df_test(table, dialect)

        table = [["A,0", "B", "C"], [1, 2, 3], [4, 5, 6]]
        dialect = SimpleDialect(delimiter=",", quotechar='"', escapechar="")
        with self.subTest(name="quoted"):
            self._df_test(table, dialect)

        table = [['a"A,0"b', "B", "C"], [1, 2, 3], [4, 5, 6]]
        dialect = SimpleDialect(delimiter=",", quotechar='"', escapechar="")
        with self.subTest(name="double"):
            self._df_test(table, dialect)

        table = [["A", "B", "C"], [1, 2, 3], [4, 5, 6]]
        dialect = SimpleDialect(delimiter=";", quotechar="", escapechar="")
        with self.subTest(name="simple_nchar"):
            self._df_test(table, dialect, num_char=10)

        table = [["Ä", "Ð", "Ç"], [1, 2, 3], [4, 5, 6]]
        dialect = SimpleDialect(delimiter=";", quotechar="", escapechar="")
        with self.subTest(name="simple_encoding"):
            self._df_test(table, dialect, num_char=10, encoding="latin1")

    def test_read_table(self) -> None:
        table: List[List[Any]]

        table = [["A", "B", "C"], [1, 2, 3], [4, 5, 6]]
        dialect = SimpleDialect(delimiter=";", quotechar="", escapechar="")
        with self.subTest(name="simple"):
            self._read_test(table, dialect)

        table = [["A,0", "B", "C"], [1, 2, 3], [4, 5, 6]]
        dialect = SimpleDialect(delimiter=",", quotechar="", escapechar="\\")
        with self.subTest(name="escaped"):
            self._read_test(table, dialect)

        table = [["A,0", "B", "C"], [1, 2, 3], [4, 5, 6]]
        dialect = SimpleDialect(delimiter=",", quotechar='"', escapechar="")
        with self.subTest(name="quoted"):
            self._read_test(table, dialect)

        table = [['a"A,0"b', "B", "C"], [1, 2, 3], [4, 5, 6]]
        dialect = SimpleDialect(delimiter=",", quotechar='"', escapechar="")
        with self.subTest(name="double"):
            self._read_test(table, dialect)

        rows = ['1,"AA"', '2,"BB"', '3,"CC"']
        exp = [["1", "AA"], ["2", "BB"], ["3", "CC"]]
        with self.subTest(name="rowtest"):
            self._read_test_rows(rows, exp)

        # This raises a NoDetectionResult due to the spacing after the
        # delimiter, which confuses the detection algorithm. Support for
        # detecting 'skipinitialspace' should fix this problem.
        rows = ['1, "AA"', '2, "BB"', '3, "CC"']
        exp = [["1", "AA"], ["2", "BB"], ["3", "CC"]]
        with self.subTest(name="raises2"):
            with self.assertRaises(NoDetectionResult):
                self._read_test_rows(rows, exp)

    def test_stream_table(self) -> None:
        table: List[List[Any]]

        table = [["A", "B", "C"], [1, 2, 3], [4, 5, 6]]
        dialect = SimpleDialect(delimiter=";", quotechar="", escapechar="")
        with self.subTest(name="simple"):
            self._stream_test(table, dialect)

        table = [["A,0", "B", "C"], [1, 2, 3], [4, 5, 6]]
        dialect = SimpleDialect(delimiter=",", quotechar="", escapechar="\\")
        with self.subTest(name="escaped"):
            self._stream_test(table, dialect)

        table = [["A,0", "B", "C"], [1, 2, 3], [4, 5, 6]]
        dialect = SimpleDialect(delimiter=",", quotechar='"', escapechar="")
        with self.subTest(name="quoted"):
            self._stream_test(table, dialect)

        table = [['a"A,0"b', "B", "C"], [1, 2, 3], [4, 5, 6]]
        dialect = SimpleDialect(delimiter=",", quotechar='"', escapechar="")
        with self.subTest(name="double"):
            self._stream_test(table, dialect)

        rows = ['1,"AA"', '2,"BB"', '3,"CC"']
        exp = [["1", "AA"], ["2", "BB"], ["3", "CC"]]
        with self.subTest(name="rowtest"):
            self._stream_test_rows(rows, exp)

        # This raises a NoDetectionResult due to the spacing after the
        # delimiter, which confuses the detection algorithm. Support for
        # detecting 'skipinitialspace' should fix this problem.
        rows = ['1, "AA"', '2, "BB"', '3, "CC"']
        exp = [["1", "AA"], ["2", "BB"], ["3", "CC"]]
        with self.subTest(name="raises2"):
            with self.assertRaises(NoDetectionResult):
                self._stream_test_rows(rows, exp)

    def _write_test_table(
        self, table: Iterable[Iterable[Any]], expected: str, **kwargs: Any
    ) -> None:
        tmpfd, tmpfname = tempfile.mkstemp(prefix="ccsv_", suffix=".csv")
        wrappers.write_table(table, tmpfname, **kwargs)
        read_encoding = kwargs.get("encoding", None)
        with open(tmpfname, "r", newline="", encoding=read_encoding) as fp:
            data = fp.read()

        try:
            self.assertEqual(data, expected)
        finally:
            os.close(tmpfd)
            os.unlink(tmpfname)

    def test_write_table(self) -> None:
        table: List[List[Union[str, int]]] = [
            ["A", "B,C", "D"],
            [1, 2, 3],
            [4, 5, 6],
        ]
        exp = 'A,"B,C",D\r\n1,2,3\r\n4,5,6\r\n'
        with self.subTest(name="default"):
            self._write_test_table(table, exp)

        dialect = SimpleDialect(delimiter=";", quotechar="", escapechar="")
        exp = "A;B,C;D\n1;2;3\n4;5;6\n"
        with self.subTest(name="dialect"):
            self._write_test_table(table, exp, dialect=dialect)

        exp = "A;1;4\nB,C;2;5\nD;3;6\n"
        with self.subTest(name="transposed"):
            self._write_test_table(table, exp, dialect=dialect, transpose=True)

        table[2].append(8)
        with self.assertRaises(ValueError):
            self._write_test_table(table, "")

        table = [["Å", "B", "C"], [1, 2, 3], [4, 5, 6]]
        exp = "Å,B,C\r\n1,2,3\r\n4,5,6\r\n"
        with self.subTest(name="encoding_1"):
            # Not specifying an encoding here could potentially fail on
            # Windows, due to open() defaulting to
            # locale.getpreferredencoding() (see gh-27).
            self._write_test_table(table, exp, encoding="utf-8")

        with self.subTest(name="encoding_2"):
            self._write_test_table(table, exp, encoding="cp1252")

    def _write_test_dicts(
        self, items: Iterable[Dict[str, Any]], expected: str, **kwargs: Any
    ) -> None:
        tmpfd, tmpfname = tempfile.mkstemp(prefix="ccsv_", suffix=".csv")
        wrappers.write_dicts(items, tmpfname, **kwargs)
        read_encoding = kwargs.get("encoding", None)
        with open(tmpfname, "r", newline="", encoding=read_encoding) as fp:
            data = fp.read()

        try:
            self.assertEqual(data, expected)
        finally:
            os.close(tmpfd)
            os.unlink(tmpfname)

    def test_write_dicts(self) -> None:
        items = [{"A": 1, "B": 2, "C": 3}, {"A": 4, "B": 5, "C": 6}]
        exp = "A,B,C\r\n1,2,3\r\n4,5,6\r\n"
        with self.subTest(name="default"):
            self._write_test_dicts(items, exp)

        dialect = SimpleDialect(delimiter=";", quotechar="", escapechar="")
        exp = "A;B;C\n1;2;3\n4;5;6\n"
        with self.subTest(name="dialect"):
            self._write_test_dicts(items, exp, dialect=dialect)

        items = [{"Å": 1, "B": 2, "C": 3}, {"Å": 4, "B": 5, "C": 6}]
        exp = "Å,B,C\r\n1,2,3\r\n4,5,6\r\n"
        with self.subTest(name="encoding_1"):
            # Not specifying an encoding here could potentially fail on
            # Windows, due to open() defaulting to
            # locale.getpreferredencoding() (see gh-27).
            self._write_test_dicts(items, exp, encoding="utf-8")

        with self.subTest(name="encoding_2"):
            self._write_test_dicts(items, exp, encoding="cp1252")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Definitions for the dialect object.

Author: Gertjan van den Burg

"""

import csv
import functools
import json

from typing import Any
from typing import Dict
from typing import Optional
from typing import Tuple
from typing import Type
from typing import Union

excel = csv.excel
excel_tab = csv.excel_tab
unix_dialect = csv.unix_dialect


@functools.total_ordering
class SimpleDialect:
    """
    The simplified dialect object.

    For the delimiter, quotechar, and escapechar the empty string means no
    delimiter/quotechar/escapechar in the file. None is used to mark it
    undefined.

    Parameters
    ----------
    delimiter : str
        The delimiter of the CSV file.

    quotechar : str
        The quotechar of the file.

    escapechar : str
        The escapechar of the file.

    strict : bool
        Whether strict parsing should be enforced. Same as in the csv module.

    """

    def __init__(
        self,
        delimiter: Optional[str],
        quotechar: Optional[str],
        escapechar: Optional[str],
        strict: bool = False,
    ):
        self.delimiter = delimiter
        self.quotechar = quotechar
        self.escapechar = escapechar
        self.strict = strict

    def validate(self) -> None:
        if self.delimiter is None or len(self.delimiter) > 1:
            raise ValueError(
                "Delimiter should be zero or one characters, got: %r"
                % self.delimiter
            )
        if self.quotechar is None or len(self.quotechar) > 1:
            raise ValueError(
                "Quotechar should be zero or one characters, got: %r"
                % self.quotechar
            )
        if self.escapechar is None or len(self.escapechar) > 1:
            raise ValueError(
                "Escapechar should be zero or one characters, got: %r"
                % self.escapechar
            )
        if self.strict not in set([False, True]):
            raise ValueError(
                "Strict should be True or False, got: %r" % self.strict
            )

    @classmethod
    def from_dict(
        cls: Type["SimpleDialect"], d: Dict[str, Any]
    ) -> "SimpleDialect":
        dialect = cls(
            d["delimiter"], d["quotechar"], d["escapechar"], strict=d["strict"]
        )
        return dialect

    @classmethod
    def from_csv_dialect(
        cls: Type["SimpleDialect"], d: csv.Dialect
    ) -> "SimpleDialect":
        delimiter = "" if d.delimiter is None else d.delimiter
        quotechar = "" if d.quoting == csv.QUOTE_NONE else d.quotechar
        escapechar = "" if d.escapechar is None else d.escapechar
        return cls(delimiter, quotechar, escapechar, strict=d.strict)

    def to_csv_dialect(self) -> csv.Dialect:
        class dialect(csv.Dialect):
            assert self.delimiter is not None
            delimiter = self.delimiter
            quotechar = '"' if self.quotechar == "" else self.quotechar
            escapechar = None if self.escapechar == "" else self.escapechar
            doublequote = True
            quoting = (
                csv.QUOTE_NONE if self.quotechar == "" else csv.QUOTE_MINIMAL
            )
            skipinitialspace = False
            # TODO: We need to set this because it can't be None anymore in
            # recent versions of Python
            lineterminator = "\n"

        return dialect()

    def to_dict(self) -> Dict[str, Union[str, bool, None]]:
        self.validate()
        d = dict(
            delimiter=self.delimiter,
            quotechar=self.quotechar,
            escapechar=self.escapechar,
            strict=self.strict,
        )
        return d

    def serialize(self) -> str:
        """Serialize dialect to a JSON object"""
        return json.dumps(self.to_dict())

    @classmethod
    def deserialize(cls: Type["SimpleDialect"], obj: str) -> "SimpleDialect":
        """Deserialize dialect from a JSON object"""
        return cls.from_dict(json.loads(obj))

    def __repr__(self) -> str:
        return "SimpleDialect(%r, %r, %r)" % (
            self.delimiter,
            self.quotechar,
            self.escapechar,
        )

    def __key(
        self,
    ) -> Tuple[Optional[str], Optional[str], Optional[str], bool]:
        return (self.delimiter, self.quotechar, self.escapechar, self.strict)

    def __hash__(self) -> int:
        return hash(self.__key())

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, SimpleDialect):
            return False
        return self.__key() == other.__key()

    def __lt__(self, other: Any) -> bool:
        # This provides a partial order on dialect objects with the goal of
        # speeding up the consistency measure.
        if not isinstance(other, SimpleDialect):
            return False
        if self.delimiter == "," and not other.delimiter == ",":
            return True
        elif other.delimiter == "," and not self.delimiter == ",":
            return False
        if self.delimiter == ";" and not other.delimiter == ";":
            return True
        elif other.delimiter == ";" and not self.delimiter == ";":
            return False
        return self.__key() < other.__key()



from .regex import *

# Names in __all__ with no definition:
#   A
#   ASCII
#   B
#   BESTMATCH
#   D
#   DEBUG
#   DEFAULT_VERSION
#   DOTALL
#   E
#   ENHANCEMATCH
#   F
#   FULLCASE
#   I
#   IGNORECASE
#   L
#   LOCALE
#   M
#   MULTILINE
#   Match
#   P
#   POSIX
#   Pattern
#   R
#   REVERSE
#   Regex
#   S
#   Scanner
#   T
#   TEMPLATE
#   U
#   UNICODE
#   V0
#   V1
#   VERBOSE
#   VERSION0
#   VERSION1
#   W
#   WORD
#   X
#   __doc__
#   __version__
#   cache_all
#   compile
#   error
#   escape
#   findall
#   finditer
#   fullmatch
#   match
#   purge
#   search
#   split
#   splititer
#   sub
#   subf
#   subfn
#   subn
#   template



# -*- coding: utf-8 -*-

"""
Integration tests for dialect detection.

Author: G.J.J. van den Burg

"""

import argparse
import gzip
import json
import multiprocessing
import os
import time
import warnings

from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple

import chardet
import termcolor

import clevercsv

from clevercsv.dialect import SimpleDialect

THIS_DIR = os.path.abspath(os.path.dirname(__file__))
SOURCE_DIR = os.path.join(THIS_DIR, "data")
TEST_FILES = os.path.join(SOURCE_DIR, "files")
TEST_DIALECTS = os.path.join(SOURCE_DIR, "dialects")

LOG_SUCCESS = os.path.join(THIS_DIR, "success.log")
LOG_ERROR = os.path.join(THIS_DIR, "error.log")
LOG_FAILED = os.path.join(THIS_DIR, "failed.log")
LOG_METHOD = os.path.join(THIS_DIR, "method.log")
LOG_RUNTIME = os.path.join(THIS_DIR, "runtime.log")

LOG_SUCCESS_PARTIAL = os.path.join(THIS_DIR, "success_partial.log")
LOG_ERROR_PARTIAL = os.path.join(THIS_DIR, "error_partial.log")
LOG_FAILED_PARTIAL = os.path.join(THIS_DIR, "failed_partial.log")
LOG_METHOD_PARTIAL = os.path.join(THIS_DIR, "method_partial.log")
LOG_RUNTIME_PARTIAL = os.path.join(THIS_DIR, "runtime_partial.log")

TIMEOUT = 5 * 60
N_BYTES_PARTIAL = 10000


def log_result(name: str, kind: str, verbose: int, partial: bool) -> None:
    table = {
        "error": (LOG_ERROR, LOG_ERROR_PARTIAL, "yellow"),
        "success": (LOG_SUCCESS, LOG_SUCCESS_PARTIAL, "green"),
        "failure": (LOG_FAILED, LOG_FAILED_PARTIAL, "red"),
    }
    assert kind in table
    outfull, outpartial, color = table[kind]
    fname = outpartial if partial else outfull

    with open(fname, "a") as fp:
        fp.write(name + "\n")
    if verbose:
        termcolor.cprint(name, color=color)


def log_method(name: str, method: str, partial: bool) -> None:
    fname = LOG_METHOD_PARTIAL if partial else LOG_METHOD
    with open(fname, "a") as fp:
        fp.write(f"{name},{method}\n")


def log_runtime(name: str, runtime: float, partial: bool) -> None:
    fname = LOG_RUNTIME_PARTIAL if partial else LOG_RUNTIME
    with open(fname, "a") as fp:
        fp.write(f"{name},{runtime}\n")


def worker(
    args: List[Any], return_dict: Dict[str, Any], **kwargs: Any
) -> None:
    det = clevercsv.Detector()
    filename, encoding, partial = args
    return_dict["error"] = False
    return_dict["dialect"] = None
    return_dict["method"] = None
    return_dict["runtime"] = float("nan")
    with gzip.open(filename, "rt", newline="", encoding=encoding) as fp:
        data = fp.read(N_BYTES_PARTIAL) if partial else fp.read()
        try:
            t = time.time()
            return_dict["dialect"] = det.detect(data, **kwargs)
            return_dict["runtime"] = time.time() - t
            return_dict["method"] = det.method_.value
        except clevercsv.Error:
            return_dict["error"] = True


def run_with_timeout(
    args: Tuple[Any, ...], kwargs: Dict[str, Any], limit: Optional[int]
) -> Tuple[Optional[SimpleDialect], bool, Optional[str], float]:
    manager = multiprocessing.Manager()
    return_dict = manager.dict()
    p = multiprocessing.Process(
        target=worker, args=(args, return_dict), kwargs=kwargs
    )
    p.start()
    p.join(limit)
    if p.is_alive():
        p.terminate()
        return None, True, None, float("nan")
    return (
        return_dict["dialect"],
        return_dict["error"],
        return_dict["method"],
        return_dict["runtime"],
    )


def run_test(
    name: str,
    gz_filename: str,
    annotation: Dict[str, Any],
    verbose: int = 1,
    partial: bool = False,
) -> None:
    if "encoding" in annotation:
        enc = annotation["encoding"]
    else:
        with gzip.open(gz_filename, "rb") as fid:
            enc = chardet.detect(fid.read())["encoding"]

    true_dialect = annotation["dialect"]
    dialect, error, method, runtime = run_with_timeout(
        (gz_filename, enc, partial), {"verbose": verbose > 1}, TIMEOUT
    )
    if error:
        return log_result(name, "error", verbose, partial)

    if dialect is None:
        log_result(name, "failure", verbose, partial)
    elif dialect.delimiter != true_dialect["delimiter"]:
        log_result(name, "failure", verbose, partial)
    elif dialect.quotechar != true_dialect["quotechar"]:
        log_result(name, "failure", verbose, partial)
    elif dialect.escapechar != true_dialect["escapechar"]:
        log_result(name, "failure", verbose, partial)
    else:
        log_result(name, "success", verbose, partial)

    assert method is not None
    log_method(name, method, partial)
    log_runtime(name, runtime, partial)


def load_test_cases() -> List[Tuple[str, str, Dict[str, Any]]]:
    cases = []
    for f in sorted(os.listdir(TEST_FILES)):
        base = f[: -len(".csv.gz")]
        dialect_file = os.path.join(TEST_DIALECTS, base + ".json")
        if not os.path.exists(dialect_file):
            continue
        filename = os.path.join(TEST_FILES, f)
        with open(dialect_file, "r") as fid:
            annotation = json.load(fid)
        if not annotation["filename"] == f[: -len(".gz")]:
            warnings.warn(
                "filename doesn't match! Input file: %s\nDialect file: %s"
                % (filename, dialect_file)
            )
            continue
        if annotation["status"] == "skip":
            continue
        cases.append((base, filename, annotation))
    return cases


def clear_output_files(partial: bool) -> None:
    files = {
        True: [
            LOG_SUCCESS_PARTIAL,
            LOG_FAILED_PARTIAL,
            LOG_ERROR_PARTIAL,
            LOG_METHOD_PARTIAL,
            LOG_RUNTIME_PARTIAL,
        ],
        False: [LOG_SUCCESS, LOG_FAILED, LOG_ERROR, LOG_METHOD, LOG_RUNTIME],
    }
    for filename in files[partial]:
        if os.path.exists(filename):
            os.unlink(filename)


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--partial",
        help="Run test with partial file data",
        action="store_true",
    )
    parser.add_argument("-v", "--verbose", help="Be verbose", action="count")
    return parser.parse_args()


def main() -> None:
    args = parse_args()
    clear_output_files(args.partial)
    cases = load_test_cases()
    for name, gz_filename, annotation in cases:
        run_test(
            name,
            gz_filename,
            annotation,
            verbose=args.verbose,
            partial=args.partial,
        )


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

"""Unit tests for encoding detection

Author: G.J.J. van den Burg
License: See the LICENSE file.

This file is part of CleverCSV.

"""

import os
import tempfile
import unittest

from dataclasses import dataclass

from typing import Any
from typing import List

from clevercsv._optional import import_optional_dependency
from clevercsv._types import AnyPath
from clevercsv.encoding import get_encoding
from clevercsv.write import writer


class EncodingTestCase(unittest.TestCase):
    @dataclass
    class Instance:
        table: List[List[Any]]
        encoding: str
        cchardet_encoding: str

    cases: List[Instance] = [
        Instance(
            table=[["Å", "B", "C"], [1, 2, 3], [4, 5, 6]],
            encoding="ISO-8859-1",
            cchardet_encoding="WINDOWS-1252",
        ),
        Instance(
            table=[["A", "B", "C"], [1, 2, 3], [4, 5, 6]],
            encoding="ascii",
            cchardet_encoding="ASCII",
        ),
        Instance(
            table=[["亜唖", "娃阿", "哀愛"], [1, 2, 3], ["挨", "姶", "葵"]],
            encoding="ISO-2022-JP",
            cchardet_encoding="ISO-2022-JP",
        ),
    ]

    def setUp(self) -> None:
        self._tmpfiles: List[AnyPath] = []

    def tearDown(self) -> None:
        for f in self._tmpfiles:
            os.unlink(f)

    def _build_file(self, table: List[List[str]], encoding: str) -> str:
        tmpfd, tmpfname = tempfile.mkstemp(
            prefix="ccsv_",
            suffix=".csv",
        )
        tmpfp = os.fdopen(tmpfd, "w", newline=None, encoding=encoding)
        w = writer(tmpfp, dialect="excel")
        w.writerows(table)
        tmpfp.close()
        self._tmpfiles.append(tmpfname)
        return tmpfname

    def test_encoding_chardet(self) -> None:
        for case in self.cases:
            table = case.table
            encoding = case.encoding
            with self.subTest(encoding=encoding):
                tmpfname = self._build_file(table, encoding)
                detected = get_encoding(tmpfname, try_cchardet=False)
                self.assertEqual(encoding, detected)

    def test_encoding_cchardet(self) -> None:
        try:
            _ = import_optional_dependency("cchardet")
        except ImportError:
            self.skipTest("Failed to import cchardet, skipping this test")

        for case in self.cases:
            table = case.table
            encoding = case.encoding
            with self.subTest(encoding=encoding):
                out_encoding = case.cchardet_encoding
                tmpfname = self._build_file(table, encoding)
                detected = get_encoding(tmpfname, try_cchardet=True)
                self.assertEqual(out_encoding, detected)


if __name__ == "__main__":
    unittest.main()
